# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import hashlib
import logging
import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest

from rsciio import IO_PLUGINS

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
pytest.importorskip("h5py", reason="h5py not installed")

from hyperspy.axes import DataAxis  # noqa: E402

TEST_DATA_PATH = Path(__file__).parent / "data"
FULLFILENAME = Path(__file__).parent / "test_io_overwriting.hspy"


class TestIOOverwriting:
    def setup_method(self, method):
        self.s = hs.signals.Signal1D(np.arange(10))
        self.new_s = hs.signals.Signal1D(np.ones(5))
        # make sure we start from a clean state
        self._clean_file()
        self.s.save(FULLFILENAME)
        self.s_file_hashed = self._hash_file(FULLFILENAME)

    def _hash_file(self, filename):
        with open(filename, "rb") as file:
            md5_hash = hashlib.md5(file.read())
            file_hashed = md5_hash.hexdigest()
        return file_hashed

    def _clean_file(self):
        if FULLFILENAME.exists():
            FULLFILENAME.unlink()

    def _check_file_is_written(self, filename):
        # Check that we have a different hash, in case the file have different
        # content from the original, the hash will be different.
        return not self.s_file_hashed == self._hash_file(filename)

    def test_io_overwriting_True(self):
        # Overwrite is True, when file exists we overwrite
        self.new_s.save(FULLFILENAME, overwrite=True)
        assert self._check_file_is_written(FULLFILENAME)

    def test_io_overwriting_False(self):
        # Overwrite if False, file exists we don't overwrite
        self.new_s.save(FULLFILENAME, overwrite=False)
        assert not self._check_file_is_written(FULLFILENAME)

    @pytest.mark.parametrize("overwrite", [None, True, False])
    def test_io_overwriting_no_existing_file(self, overwrite):
        self._clean_file()  # remove the file
        self.new_s.save(FULLFILENAME, overwrite=overwrite)
        assert self._check_file_is_written(FULLFILENAME)

    def test_io_overwriting_None_existing_file_y(self):
        # Overwrite is None, when file exists we ask, mock `y` here
        with patch("builtins.input", return_value="y"):
            self.new_s.save(FULLFILENAME)
            assert self._check_file_is_written(FULLFILENAME)

    def test_io_overwriting_None_existing_file_n(self):
        # Overwrite is None, when file exists we ask, mock `n` here
        with patch("builtins.input", return_value="n"):
            self.new_s.save(FULLFILENAME)
            assert not self._check_file_is_written(FULLFILENAME)

    def test_io_overwriting_invalid_parameter(self):
        with pytest.raises(ValueError, match="parameter can only be"):
            self.new_s.save(FULLFILENAME, overwrite="spam")

    def teardown_method(self, method):
        self._clean_file()


class TestNonUniformAxisCheck:
    def setup_method(self, method):
        axis = DataAxis(axis=1 / (np.arange(10) + 1), navigate=False)
        self.s = hs.signals.Signal1D(np.arange(10), axes=(axis.get_axis_dictionary(),))
        # make sure we start from a clean state

    def test_io_nonuniform(self, tmp_path):
        assert self.s.axes_manager[0].is_uniform is False
        self.s.save(tmp_path / "tmp.hspy")
        with pytest.raises(TypeError, match="not supported for non-uniform"):
            self.s.save(tmp_path / "tmp.msa")

    def test_nonuniform_writer_characteristic(self):
        for plugin in IO_PLUGINS:
            if "non_uniform_axis" not in plugin:
                print(  # noqa: T201
                    f"{plugin.name} IO-plugin is missing the "
                    "characteristic `non_uniform_axis`"
                )

    def test_nonuniform_error(self, tmp_path):
        assert self.s.axes_manager[0].is_uniform is False
        incompatible_writers = [
            plugin["file_extensions"][plugin["default_extension"]]
            for plugin in IO_PLUGINS
            if (
                plugin["writes"] is True
                or plugin["writes"] is not False
                and [1, 0] in plugin["writes"]
            )
            and not plugin["non_uniform_axis"]
        ]
        for ext in incompatible_writers:
            with pytest.raises(TypeError, match="not supported for non-uniform"):
                filename = "tmp." + ext
                self.s.save(tmp_path / filename, overwrite=True)


def test_glob_wildcards():
    s = hs.signals.Signal1D(np.arange(10))

    with tempfile.TemporaryDirectory() as dirpath:
        fnames = [os.path.join(dirpath, f"temp[1x{x}].hspy") for x in range(2)]

        for f in fnames:
            s.save(f)

        with pytest.raises(ValueError, match="No filename matches the pattern"):
            _ = hs.load(fnames[0])

        t = hs.load([fnames[0]])
        assert len(t) == 1

        t = hs.load(fnames)
        assert len(t) == 2

        t = hs.load(os.path.join(dirpath, "temp*.hspy"))
        assert len(t) == 2

        t = hs.load(
            os.path.join(dirpath, "temp[*].hspy"),
            escape_square_brackets=True,
        )
        assert len(t) == 2

        with pytest.raises(ValueError, match="No filename matches the pattern"):
            _ = hs.load(os.path.join(dirpath, "temp[*].hspy"))

        # Test pathlib.Path
        t = hs.load(Path(dirpath, "temp[1x0].hspy"))
        assert len(t) == 1

        t = hs.load([Path(dirpath, "temp[1x0].hspy"), Path(dirpath, "temp[1x1].hspy")])
        assert len(t) == 2

        t = hs.load(list(Path(dirpath).glob("temp*.hspy")))
        assert len(t) == 2

        t = hs.load(Path(dirpath).glob("temp*.hspy"))
        assert len(t) == 2


def test_file_not_found_error(tmp_path):
    temp_fname = tmp_path / "temp.hspy"

    if os.path.exists(temp_fname):
        os.remove(temp_fname)

    with pytest.raises(ValueError, match="No filename matches the pattern"):
        _ = hs.load(temp_fname)

    with pytest.raises(FileNotFoundError):
        _ = hs.load([temp_fname])


def test_file_reader_error(tmp_path):
    # Only None, str or objects with attr "file_reader" are supported
    s = hs.signals.Signal1D(np.arange(10))

    f = tmp_path / "temp.hspy"
    s.save(f)

    with pytest.raises(ValueError, match="reader"):
        _ = hs.load(f, reader=123)


def test_file_reader_warning(caplog, tmp_path):
    s = hs.signals.Signal1D(np.arange(10))

    f = tmp_path / "temp.hspy"
    s.save(f)
    try:
        with caplog.at_level(logging.WARNING):
            _ = hs.load(f, reader="some_unknown_file_extension")

        assert "Unable to infer file type from extension" in caplog.text
    except (ValueError, OSError):
        # Test fallback to Pillow imaging library
        pass


def test_file_reader_options():
    s = hs.signals.Signal1D(np.arange(10))

    with tempfile.TemporaryDirectory() as dirpath:
        f = os.path.join(dirpath, "temp.hspy")
        s.save(f)
        f2 = os.path.join(dirpath, "temp.emd")
        s.save(f2)

        # Test string reader
        t = hs.load(Path(dirpath, "temp.hspy"), reader="hspy")
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))

        # Test string reader uppercase
        t = hs.load(Path(dirpath, "temp.hspy"), reader="HSpy")
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))

        # Test string reader alias
        t = hs.load(Path(dirpath, "temp.hspy"), reader="hyperspy")
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))

        # Test string reader name
        t = hs.load(Path(dirpath, "temp.emd"), reader="emd")
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))

        # Test string reader aliases
        t = hs.load(Path(dirpath, "temp.emd"), reader="Electron Microscopy Data (EMD)")
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))
        t = hs.load(Path(dirpath, "temp.emd"), reader="Electron Microscopy Data")
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))

        # Test object reader
        from rsciio import hspy

        t = hs.load(Path(dirpath, "temp.hspy"), reader=hspy)
        assert len(t) == 1
        np.testing.assert_allclose(t.data, np.arange(10))


def test_save_default_format(tmp_path):
    s = hs.signals.Signal1D(np.arange(10))

    f = tmp_path / "temp"
    s.save(f)

    t = hs.load(tmp_path / "temp.hspy")
    assert len(t) == 1


def test_load_original_metadata(tmp_path):
    s = hs.signals.Signal1D(np.arange(10))
    s.original_metadata.a = 0

    f = tmp_path / "temp"
    s.save(f)
    assert s.original_metadata.as_dictionary() != {}

    t = hs.load(tmp_path / "temp.hspy")
    assert t.original_metadata.as_dictionary() == s.original_metadata.as_dictionary()

    t = hs.load(tmp_path / "temp.hspy", load_original_metadata=False)
    assert t.original_metadata.as_dictionary() == {}


def test_load_save_filereader_metadata():
    # tests that original FileReader metadata is correctly persisted and
    # appended through a save and load cycle
    s = hs.load(TEST_DATA_PATH / "msa" / "example1.msa")
    assert s.metadata.General.FileIO.Number_0.io_plugin == "rsciio.msa"
    assert s.metadata.General.FileIO.Number_0.operation == "load"
    assert s.metadata.General.FileIO.Number_0.hyperspy_version == hs.__version__

    with tempfile.TemporaryDirectory() as dirpath:
        f = os.path.join(dirpath, "temp")
        s.save(f)
        expected = {
            "0": {
                "io_plugin": "rsciio.msa",
                "operation": "load",
                "hyperspy_version": hs.__version__,
            },
            "1": {
                "io_plugin": "rsciio.hspy",
                "operation": "save",
                "hyperspy_version": hs.__version__,
            },
            "2": {
                "io_plugin": "rsciio.hspy",
                "operation": "load",
                "hyperspy_version": hs.__version__,
            },
        }
        del s.metadata.General.FileIO.Number_0.timestamp  # runtime dependent
        del s.metadata.General.FileIO.Number_1.timestamp  # runtime dependent
        assert s.metadata.General.FileIO.Number_0.as_dictionary() == expected["0"]
        assert s.metadata.General.FileIO.Number_1.as_dictionary() == expected["1"]

        t = hs.load(Path(dirpath, "temp.hspy"))
        del t.metadata.General.FileIO.Number_0.timestamp  # runtime dependent
        del t.metadata.General.FileIO.Number_1.timestamp  # runtime dependent
        del t.metadata.General.FileIO.Number_2.timestamp  # runtime dependent
        assert t.metadata.General.FileIO.as_dictionary() == expected



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import gc
import shutil
import zipfile
from pathlib import Path

import dask.array as da
import numpy as np
import pytest
from dask.array.core import normalize_chunks

from rsciio.exceptions import VisibleDeprecationWarning
from rsciio.quantumdetector._api import (
    MIBProperties,
    load_mib_data,
    parse_exposures,
    parse_hdr_file,
    parse_timestamps,
)

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
zarr = pytest.importorskip("zarr", reason="zarr not installed")


TEST_DATA_DIR = Path(__file__).parent / "data" / "quantumdetector"
ZIP_FILE = TEST_DATA_DIR / "Merlin_Single_Quad.zip"
ZIP_FILE2 = TEST_DATA_DIR / "Merlin_navigation4x2_ROI.zip"
ZIP_FILE3 = TEST_DATA_DIR / "Merlin_navigation4x2_signalNx256_ROI.zip"
TEST_DATA_DIR_UNZIPPED = TEST_DATA_DIR / "unzipped"


SINGLE_CHIP_FNAME_LIST = [
    f"Single_{frame}_Frame_CounterDepth_{depth}_Rows_256.mib"
    for frame in [1, 9]
    for depth in [1, 6, 12, 24]
]


QUAD_CHIP_FNAME_LIST = [
    f"Quad_{frame}_Frame_CounterDepth_{depth}_Rows_256.mib"
    for frame in [1, 9]
    for depth in [1, 6, 12, 24]
]

SIGNAL_ROI_FNAME_LIST = [
    "002_merlin_test_roi_sig256x128_nav4x2_hot_pixel_52x_39y.mib",
    "003_merlin_test_roi_sig256x64_nav4x2_hot_pixel_52x_39y.mib",
]


def filter_list(fname_list, string):
    return [fname for fname in fname_list if string in fname]


def setup_module():
    if not TEST_DATA_DIR_UNZIPPED.exists():
        if ZIP_FILE.exists():
            with zipfile.ZipFile(ZIP_FILE, "r") as zipped:
                zipped.extractall(TEST_DATA_DIR_UNZIPPED)

        if ZIP_FILE2.exists():
            with zipfile.ZipFile(ZIP_FILE2, "r") as zipped:
                zipped.extractall(TEST_DATA_DIR_UNZIPPED)
        if ZIP_FILE3.exists():
            with zipfile.ZipFile(ZIP_FILE3, "r") as zipped:
                zipped.extractall(TEST_DATA_DIR_UNZIPPED)


def teardown_module():
    # necessary on windows, to help closing the files...
    gc.collect()
    shutil.rmtree(TEST_DATA_DIR_UNZIPPED)


def _get_expected_dtype_from_fname(fname):
    counter_depth = int(fname.split("CounterDepth_")[1].split("_Rows")[0])
    if counter_depth in [1, 6]:
        dtype = np.dtype(">u1")
    elif counter_depth == 12:
        dtype = np.dtype(">u2")
    else:
        dtype = np.dtype(">u4")
    return dtype


@pytest.mark.parametrize(
    ("fname", "reshape"),
    zip(
        SINGLE_CHIP_FNAME_LIST + filter_list(SINGLE_CHIP_FNAME_LIST, "9_Frames"),
        [False] * len(SINGLE_CHIP_FNAME_LIST)
        + [True] * len(filter_list(SINGLE_CHIP_FNAME_LIST, "9_Frames")),
    ),
)
def test_single_chip(fname, reshape):
    if "9_Frame" in fname:
        navigation_shape = (3, 3) if reshape else (9,)
    else:
        navigation_shape = ()

    nav_shape = navigation_shape if reshape else None
    s = hs.load(TEST_DATA_DIR_UNZIPPED / fname, navigation_shape=nav_shape)
    assert s.data.shape == navigation_shape + (256, 256)
    assert s.data.dtype == _get_expected_dtype_from_fname(fname)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == navigation_shape

    for axis in s.axes_manager.signal_axes:
        assert axis.scale == 1
        assert axis.offset == 0
        assert axis.units == ""


@pytest.mark.parametrize("fname", QUAD_CHIP_FNAME_LIST)
def test_quad_chip(fname):
    s = hs.load(TEST_DATA_DIR_UNZIPPED / fname)
    if "9_Frame" in fname:
        if "24_Rows_256" in fname:
            # Unknown why the timestamps of this file are not consistent
            # with others
            navigation_shape = (3, 3)
        else:
            navigation_shape = (9,)
    else:
        navigation_shape = ()
    assert s.data.shape == navigation_shape + (512, 512)
    assert s.data.dtype == _get_expected_dtype_from_fname(fname)
    assert s.axes_manager.signal_shape == (512, 512)
    assert s.axes_manager.navigation_shape == navigation_shape

    for axis in s.axes_manager.signal_axes:
        assert axis.scale == 1
        assert axis.offset == 0
        assert axis.units == ""


@pytest.mark.parametrize(
    "chunks", ("auto", (3, 3, 128, 128), ("auto", "auto", 128, 128))
)
def test_chunks(chunks):
    fname = TEST_DATA_DIR_UNZIPPED / "Quad_9_Frame_CounterDepth_24_Rows_256.mib"
    s = hs.load(fname, lazy=True, chunks=chunks)
    chunks = normalize_chunks(chunks, shape=s.data.shape, dtype=s.data.dtype)
    assert s.data.chunks == chunks


def test_mib_properties_single__repr__():
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    mib_prop = MIBProperties()
    mib_prop.parse_file(str(fname))
    assert "\nPath: " == mib_prop.__repr__()[:7]


def test_mib_properties_quad__repr__():
    fname = TEST_DATA_DIR_UNZIPPED / "Quad_9_Frame_CounterDepth_1_Rows_256.mib"
    mib_prop = MIBProperties()
    mib_prop.parse_file(str(fname))
    assert "\nPath: " == mib_prop.__repr__()[:7]


def test_interrupted_acquisition():
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    # There is only 9 frames, simulate interrupted acquisition using 10 lines
    s = hs.load(fname, navigation_shape=(4, 3))
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (4, 2)

    s = hs.load(TEST_DATA_DIR_UNZIPPED / fname, navigation_shape=(2, 4))
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (2, 4)


def test_interrupted_acquisition_first_frame():
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    # There is only 9 frames, simulate interrupted acquisition using 10 lines
    s = hs.load(fname, navigation_shape=(10, 2), first_frame=1)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (8,)

    s = hs.load(fname, navigation_shape=(10, 2), first_frame=2)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (7,)


@pytest.mark.parametrize("navigation_shape", (None, (8,), (4, 2)))
def test_non_square(navigation_shape):
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"
    s = hs.load(fname, navigation_shape=navigation_shape)
    assert s.axes_manager.signal_shape == (256, 256)
    if navigation_shape is None:
        navigation_shape = (4, 2)
    assert s.axes_manager.navigation_shape == navigation_shape


def test_no_hdr():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"
    fname2 = str(fname).replace(".mib", "-copy.mib")
    shutil.copyfile(fname, fname2)
    s = hs.load(fname2)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (4, 2)


@pytest.mark.parametrize(
    "kwargs",
    (
        {"first_frame": None, "last_frame": None},
        {"first_frame": 0, "last_frame": 9},
        {"first_frame": -9, "last_frame": 9},
        {"first_frame": -9},
        {"first_frame": 0},
        {"last_frame": None},
        {"last_frame": 9},
    ),
)
def test_first_last_frame_all9(kwargs):
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    s = hs.load(fname, **kwargs)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (9,)
    assert s.data.shape == (9, 256, 256)


@pytest.mark.parametrize("navigation_shape", ((8,), (4, 2)))
@pytest.mark.parametrize(
    "kwargs",
    (
        {"first_frame": 0, "last_frame": -1},
        {"first_frame": 0, "last_frame": 8},
        {"first_frame": 1, "last_frame": 9},
        {"first_frame": -8, "last_frame": 9},
        {"first_frame": 1},
        {"last_frame": -1},
        {"last_frame": 8},
    ),
)
def test_first_last_frame_8(kwargs, navigation_shape):
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    s = hs.load(fname, navigation_shape=navigation_shape, **kwargs)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == navigation_shape
    assert s.data.shape == navigation_shape[::-1] + (256, 256)


def test_first_last_frame_8_nav_shape_None():
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    # the navigation_shape will be obtained from the hdf file
    s = hs.load(fname, navigation_shape=None, first_frame=None, last_frame=-1)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (8,)
    assert s.data.shape == (8, 256, 256)

    s = hs.load(fname, navigation_shape=None, first_frame=1, last_frame=None)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (8,)
    assert s.data.shape == (8, 256, 256)


@pytest.mark.parametrize("return_mmap", (True, False))
@pytest.mark.parametrize("lazy", (True, False))
def test_load_mib_data(lazy, return_mmap):
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"
    data = load_mib_data(str(fname), lazy=lazy, return_mmap=return_mmap)
    assert data.shape == (8, 256, 256)
    if return_mmap or not lazy:
        # Even when lazy, it should still be an instance of
        # np.ndarray because it should return the memmap
        assert isinstance(data, np.ndarray)
    else:
        assert isinstance(data, da.Array)

    data = load_mib_data(str(fname), navigation_shape=(4, 2))
    assert data.shape == (2, 4, 256, 256)

    data, headers = load_mib_data(str(fname), return_headers=True)
    assert data.shape == (8, 256, 256)
    assert headers.shape == (8,)


@pytest.mark.parametrize("lazy", (True, False))
def test_load_mib_data_return_mmap_default(lazy):
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"
    data = load_mib_data(str(fname), lazy=lazy)
    # Even if this lazy, it should still be an instance of np.ndarray
    # because it should return the memmap
    assert isinstance(data, np.ndarray)


def test_test_load_mib_data_from_buffer():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"

    with open(fname, mode="rb") as f:
        data = load_mib_data(f.read())

    assert data.shape == (8, 256, 256)

    with open(fname, mode="rb") as f:
        data = load_mib_data(f.read(), navigation_shape=(4, 2))

    assert data.shape == (2, 4, 256, 256)

    with open(fname, mode="rb") as f:
        data, headers = load_mib_data(
            f.read(), navigation_shape=(4, 2), return_headers=True
        )

    assert data.shape == (2, 4, 256, 256)
    assert headers.shape == (2, 4)

    with open(fname, mode="rb") as f:
        with pytest.raises(ValueError):
            # loading lazy memory buffer is not supported
            _ = load_mib_data(f.read(), lazy=True)


def test_parse_exposures():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"

    data, headers = load_mib_data(str(fname), return_headers=True, return_mmap=True)
    exposures = parse_exposures(headers[0])
    assert exposures == [100.0]

    exposures = parse_exposures(headers)
    assert exposures == [100.0] * headers.shape[0]

    with pytest.raises(ValueError):
        load_mib_data(str(fname), return_headers=True, return_mmap=False)


def test_parse_timestamps():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"

    data, headers = load_mib_data(str(fname), return_headers=True, return_mmap=True)
    timestamps = parse_timestamps(headers[0])
    assert timestamps == ["2021-05-07T16:51:10.905800928Z"]

    timestamps = parse_timestamps(headers)
    assert len(timestamps) == len(headers)


def test_metadata():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"

    s = hs.load(fname)
    md_gen = s.metadata.General
    assert md_gen.date == "2021-05-07"
    assert md_gen.time == "16:51:10.905800928"
    assert md_gen.time_zone == "UTC"
    np.testing.assert_allclose(s.metadata.Acquisition_instrument.dwell_time, 1e-1)


def test_print_info():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"

    _ = hs.load(fname, print_info=True)


def test_navigation_shape_list_error():
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"

    with pytest.raises(TypeError):
        _ = hs.load(fname, navigation_shape=[4, 2])


def test_load_save_cycle(tmp_path):
    fname = TEST_DATA_DIR_UNZIPPED / "001_4x2_6bit.mib"
    s = hs.load(fname, navigation_shape=(4, 2))
    fname2 = tmp_path / "test.zspy"
    s.save(fname2)
    s2 = hs.load(fname2)

    np.testing.assert_allclose(s.data, s2.data)
    assert s.axes_manager.navigation_shape == s2.axes_manager.navigation_shape
    assert s.axes_manager.signal_shape == s2.axes_manager.signal_shape
    assert s.data.dtype == s2.data.dtype


def test_frames_in_acquisition_zero():
    # Some hdr file have entry "Frames per Trigger (Number): 0"
    # Possibly for "continuous and indefinite" acquisition
    # Copy and edit a file with corresponding changes
    base_fname = TEST_DATA_DIR_UNZIPPED / "Single_1_Frame_CounterDepth_6_Rows_256"
    fname = f"{base_fname}_zero_frames_in_acquisition"
    # Create test file using existing test file
    shutil.copyfile(f"{base_fname}.mib", f"{fname}.mib")
    hdf_dict = parse_hdr_file(f"{base_fname}.hdr")
    hdf_dict["Frames in Acquisition (Number)"] = 0
    with open(f"{fname}.hdr", "w") as f:
        f.write("HDR\n")
        for k, v in hdf_dict.items():
            f.write(f"{k}:\t{v}\n")
        f.write("End\t")

    s = hs.load(f"{fname}.mib")
    assert s.axes_manager.navigation_shape == ()


@pytest.mark.parametrize("fname", SIGNAL_ROI_FNAME_LIST)
def test_hot_pixel_signal_ROI(fname):
    s = hs.load(TEST_DATA_DIR_UNZIPPED / fname)
    for i in s:
        for j in i:
            data = j.data
            xy = np.argwhere(data == data.max())
            assert len(xy) == 1
            coord_shifted = np.array(*xy) - np.array([data.shape[0], 0])
            assert np.all(coord_shifted == np.array([-40, 52]))


@pytest.mark.parametrize("fname", SIGNAL_ROI_FNAME_LIST)
def test_signal_shape_ROI(fname):
    s = hs.load(TEST_DATA_DIR_UNZIPPED / fname)
    assert s.axes_manager.navigation_shape == (4, 2)
    if "sig256x64" in fname:
        assert s.axes_manager.signal_shape == (256, 64)
    if "sig256x128" in fname:
        assert s.axes_manager.signal_shape == (256, 128)


@pytest.mark.parametrize("distributed", [True, False])
def test_deprecated_distributed(distributed):
    fname = TEST_DATA_DIR_UNZIPPED / "Single_9_Frame_CounterDepth_1_Rows_256.mib"
    with pytest.warns(VisibleDeprecationWarning):
        _ = hs.load(fname, distributed=distributed)

    with pytest.warns(VisibleDeprecationWarning):
        load_mib_data(str(fname), distributed=distributed)



"""skimage's `rescale_intensity` that takes and returns dask arrays."""

import warnings

import numpy as np
import skimage
from packaging.version import Version
from skimage.exposure.exposure import _output_dtype, intensity_range


def rescale_intensity(image, in_range="image", out_range="dtype"):
    """Return image after stretching or shrinking its intensity levels.

    The desired intensity range of the input and output, `in_range` and
    `out_range` respectively, are used to stretch or shrink the intensity range
    of the input image. See examples below.

    Parameters
    ----------
    image : array
        Image array.
    in_range, out_range : str or 2-tuple, optional
        Min and max intensity values of input and output image.
        The possible values for this parameter are enumerated below.

        'image'
            Use image min/max as the intensity range.
        'dtype'
            Use min/max of the image's dtype as the intensity range.
        dtype-name
            Use intensity range based on desired `dtype`. Must be valid key
            in `DTYPE_RANGE`.
        2-tuple
            Use `range_values` as explicit min/max intensities.

    Returns
    -------
    out : array
        Image array after rescaling its intensity. This image is the same dtype
        as the input image.

    Notes
    -----
    .. versionchanged:: 0.17
        The dtype of the output array has changed to match the input dtype, or
        float if the output range is specified by a pair of floats.

    See Also
    --------
    equalize_hist

    Examples
    --------
    By default, the min/max intensities of the input image are stretched to
    the limits allowed by the image's dtype, since `in_range` defaults to
    'image' and `out_range` defaults to 'dtype':

    >>> image = np.array([51, 102, 153], dtype=np.uint8)
    >>> rescale_intensity(image)
    array([  0, 127, 255], dtype=uint8)

    It's easy to accidentally convert an image dtype from uint8 to float:

    >>> 1.0 * image
    array([ 51., 102., 153.])

    Use `rescale_intensity` to rescale to the proper range for float dtypes:

    >>> image_float = 1.0 * image
    >>> rescale_intensity(image_float)
    array([0. , 0.5, 1. ])

    To maintain the low contrast of the original, use the `in_range` parameter:

    >>> rescale_intensity(image_float, in_range=(0, 255))
    array([0.2, 0.4, 0.6])

    If the min/max value of `in_range` is more/less than the min/max image
    intensity, then the intensity levels are clipped:

    >>> rescale_intensity(image_float, in_range=(0, 102))
    array([0.5, 1. , 1. ])

    If you have an image with signed integers but want to rescale the image to
    just the positive range, use the `out_range` parameter. In that case, the
    output dtype will be float:

    >>> image = np.array([-10, 0, 10], dtype=np.int8)
    >>> rescale_intensity(image, out_range=(0, 127))
    array([  0. ,  63.5, 127. ])

    To get the desired range with a specific dtype, use ``.astype()``:

    >>> rescale_intensity(image, out_range=(0, 127)).astype(np.int8)
    array([  0,  63, 127], dtype=int8)

    If the input image is constant, the output will be clipped directly to the
    output range:
    >>> image = np.array([130, 130, 130], dtype=np.int32)
    >>> rescale_intensity(image, out_range=(0, 127)).astype(np.int32)
    array([127, 127, 127], dtype=int32)
    """
    args = ()
    if Version(skimage.__version__) >= Version("0.19.0"):
        args = (image.dtype,)
    if out_range in ["dtype", "image"]:
        out_dtype = _output_dtype(image.dtype.type, *args)
    else:
        out_dtype = _output_dtype(out_range, *args)

    imin, imax = map(float, intensity_range(image, in_range))
    omin, omax = map(
        float, intensity_range(image, out_range, clip_negative=(imin >= 0))
    )

    if np.any(np.isnan([imin, imax, omin, omax])):
        warnings.warn(
            "One or more intensity levels are NaN. Rescaling will broadcast "
            "NaN to the full image. Provide intensity levels yourself to "
            "avoid this. E.g. with np.nanmin(image), np.nanmax(image).",
            stacklevel=2,
        )

    image = np.clip(image, imin, imax)

    if imin != imax:
        image = (image - imin) / (imax - imin)
        return (image * (omax - omin) + omin).astype(dtype=out_dtype)
    else:
        return np.clip(image, omin, omax).astype(out_dtype)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

# The details of the format were taken from
# https://www.biochem.mpg.de/doc_tom/TOM_Release_2008/IOfun/tom_mrcread.html
# and https://ami.scripps.edu/software/mrctools/mrc_specification.php

import importlib.util
import logging
import xml.etree.ElementTree as ET
from copy import deepcopy
from pathlib import Path

import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC

_logger = logging.getLogger(__name__)


def _remove_none_from_dict(dict_in):
    for key, value in list(dict_in.items()):
        if isinstance(value, dict):
            _remove_none_from_dict(value)
        elif value is None:
            del dict_in[key]


class JobinYvonXMLReader:
    """Class to read Jobin Yvon .xml-files.

    The file is read using xml.etree.ElementTree.
    Each element can have the following attributes: attrib, tag, text.
    Moreover, non-leaf-elements are iterable (iterate over child-nodes).
    In this specific format, the tags do not contain useful information.
    Instead, the "ID"-entry in attrib is used to identify the sort of information.
    The IDs are consistent for the tested files.

    Parameters
    ----------
    file_path: pathlib.Path
        Path to the to be read file.

    use_uniform_signal_axis: bool, default=False
        Decides whether to use uniform or non-uniform signal-axis.

    Attributes
    ----------
    data, metadata, original_metadata, axes

    Methods
    -------
    parse_file, get_original_metadata, get_axes, get_data, map_metadata
    """

    def __init__(self, file_path, use_uniform_signal_axis=False):
        self._file_path = file_path
        self._use_uniform_signal_axis = use_uniform_signal_axis

        if importlib.util.find_spec("lumispy") is None:
            self._lumispy_installed = False
            _logger.warning(
                "Cannot find package lumispy, using BaseSignal1D as signal class."
            )
        else:
            self._lumispy_installed = True  # pragma: no cover

    @property
    def _signal_type(self):
        if self._lumispy_installed:
            return "Luminescence"  # pragma: no cover
        else:
            return ""

    @staticmethod
    def _get_id(xml_element):
        return xml_element.attrib["ID"]

    @staticmethod
    def _get_size(xml_element):
        return int(xml_element.attrib["Size"])

    def parse_file(self):
        """First parse through file to extract data/metadata positions."""
        tree = ET.parse(self._file_path)
        root = tree.getroot()

        lsx_tree_list = root.findall("LSX_Tree")
        if len(lsx_tree_list) > 1:
            _logger.critical(
                "File contains multiple positions to read metadata from.\n"
                "The first location is chosen."
            )  # pragma: no cover
        elif len(lsx_tree_list) == 0:
            _logger.critical("No metadata found.")  # pragma: no cover
        lsx_tree = lsx_tree_list[0]

        lsx_matrix_list = root.findall("LSX_Matrix")
        if len(lsx_matrix_list) > 1:
            _logger.critical(
                "File contains multiple positions to read data from.\n"
                "The first location is chosen."
            )  # pragma: no cover
        elif len(lsx_matrix_list) == 0:
            _logger.critical("No data found.")  # pragma: no cover
        self._lsx_matrix = lsx_matrix_list[0]

        for child in lsx_tree:
            id = self._get_id(child)
            if id == "0x6C62D4D9":
                self._metadata_root = child
            if id == "0x6C7469D9":
                self._title = child.text
            if id == "0x6D707974":
                self._measurement_type = child.text
            if id == "0x6C676EC6":
                for child2 in child:
                    if self._get_id(child2) == "0x0":
                        self._angle = child2.text
            if id == "0x7A74D9D6":
                for child2 in child:
                    if self._get_id(child2) == "0x7B697861":
                        self._axis_root = child2

        if not hasattr(self, "_metadata_root"):
            _logger.critical("Could not extract metadata")  # pragma: no cover
        if not hasattr(self, "_axis_root"):
            _logger.critical("Could not extract axis")  # pragma: no cover

    def _get_metadata_values(self, xml_element, tag):
        """Helper method to extract information from metadata xml-element.

        Parameters
        ----------
        xml_element: xml.etree.ElementTree.Element
            Head level metadata element.
        tag: Str
            Used as the corresponding key in the original_metadata dictionary.

        Examples
        --------
        <LSX Format="9" ID="0x6D7361DE" Size="5">
                        <LSX Format="7" ID="0x6D6D616E">Laser (nm)</LSX>
                        <LSX Format="7" ID="0x7D6C61DB">633 </LSX>
                        <LSX Format="5" ID="0x8736F70">632.817</LSX>
                        <LSX Format="4" ID="0x6A6DE7D3">0</LSX>
        </LSX>

        This corresponds to child-level (xml-element would be multiple of those elements).
        ID="0x6D6D616E" -> key name for original metadata
        ID="0x7D6C61DB" -> first value
        ID="0x8736F70" -> second value

        Which value is used is decided in _clean_up_metadata (in this case the second value is used).
        """
        metadata_xml_element = dict()
        for child in xml_element:
            values = {}
            for child2 in child:
                if self._get_id(child2) == "0x6D6D616E":
                    key = child2.text
                if self._get_id(child2) == "0x7D6C61DB":
                    values["1"] = child2.text
                if self._get_id(child2) == "0x8736F70":
                    values["2"] = child2.text
            metadata_xml_element[key] = values
        self.original_metadata[tag] = metadata_xml_element

    def _clean_up_metadata(self):
        """Cleans up original metadata to meet standardized format.

        This means converting numbers from strings to floats,
        deciding which value shall be used (when 2 values are extracted,
        see _get_metadata_values() for more information).
        Moreover, some names are slightly modified.
        """
        convert_to_numeric = [
            "Acq. time (s)",
            "Accumulations",
            "Delay time (s)",
            "Binning",
            "Detector temperature (°C)",
            "Objective",
            "Grating",
            "ND Filter",
            "Laser (nm)",
            "Spectro (nm)",
            "Hole",
            "Laser Pol. (°)",
            "Raman Pol. (°)",
            "X (µm)",
            "Y (µm)",
            "Z (µm)",
            "Full time(s)",
            "rotation angle (rad)",
            "Windows",
        ]

        change_to_second_value = [
            "Objective",
            "Grating",
            "ND Filter",
            "Laser (nm)",
            "Spectro (nm)",
        ]

        ## use second extracted value
        for key in change_to_second_value:
            try:
                self.original_metadata["experimental_setup"][key] = (
                    self.original_metadata["experimental_setup"][key]["2"]
                )
            except KeyError:
                pass

        ## use first extracted value
        for key, value in self.original_metadata["experimental_setup"].items():
            if isinstance(value, dict):
                # only if there is an entry/value
                if bool(value):
                    self.original_metadata["experimental_setup"][key] = (
                        self.original_metadata["experimental_setup"][key]["1"]
                    )

        for key, value in self.original_metadata["date"].items():
            if isinstance(value, dict):
                if bool(value):
                    self.original_metadata["date"][key] = self.original_metadata[
                        "date"
                    ][key]["1"]

        for key, value in self.original_metadata["file_information"].items():
            if isinstance(value, dict):
                if bool(value):
                    self.original_metadata["file_information"][key] = (
                        self.original_metadata["file_information"][key]["1"]
                    )

        ## convert strings to float
        for key in convert_to_numeric:
            try:
                self.original_metadata["experimental_setup"][key] = float(
                    self.original_metadata["experimental_setup"][key]
                )
            except KeyError:
                pass

        ## move the unit from grating to the key name
        try:
            self.original_metadata["experimental_setup"]["Grating (gr/mm)"] = (
                self.original_metadata["experimental_setup"].pop("Grating")
            )
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover

        ## add percentage for filter key name
        try:
            self.original_metadata["experimental_setup"]["ND Filter (%)"] = (
                self.original_metadata["experimental_setup"].pop("ND Filter")
            )
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover

    def get_original_metadata(self):
        """Extracts metadata from file."""
        self.original_metadata = {}
        for child in self._metadata_root:
            id = self._get_id(child)
            if id == "0x7CECDBD7":
                date = child
            if id == "0x8716361":
                metadata = child
            if id == "0x7C73E2D2":
                file_specs = child

        ## setup tree structure original_metadata -> date{...}, experimental_setup{...}, file_information{...}
        ## based on structure in file
        self._get_metadata_values(date, "date")
        self._get_metadata_values(metadata, "experimental_setup")
        self._get_metadata_values(file_specs, "file_information")
        try:
            self.original_metadata["experimental_setup"]["measurement_type"] = (
                self._measurement_type
            )
        except AttributeError:  # pragma: no cover
            pass  # pragma: no cover
        try:
            self.original_metadata["experimental_setup"]["title"] = self._title
        except AttributeError:  # pragma: no cover
            pass  # pragma: no cover
        try:
            self.original_metadata["experimental_setup"]["rotation angle (rad)"] = (
                self._angle
            )
        except AttributeError:
            pass
        self._clean_up_metadata()

    def _set_signal_type(self, xml_element):
        """Sets signal type and units based on metadata from file.

        Extra method, because this information is stored separate from the rest of the metadata.

        Parameters
        ----------
        xml_element: xml.etree.ElementTree.Element
            Head level metadata element.
        """
        for child in xml_element:
            id = self._get_id(child)
            ## contains also intensity-minima/maxima-values for each data-row (ignored by this reader)
            if id == "0x6D707974":
                self.original_metadata["experimental_setup"]["signal type"] = child.text
            if id == "0x7C696E75":
                self.original_metadata["experimental_setup"]["signal units"] = (
                    child.text
                )

    def _set_nav_axis(self, xml_element, tag):
        """Helper method for setting navigation axes.

        Parameters
        ----------
        xml_element: xml.etree.ElementTree.Element
            Head level metadata element.
        tag: Str
            Axis name.
        """
        has_nav = True
        nav_dict = dict()
        for child in xml_element:
            id = self._get_id(child)
            if id == "0x6D707974":
                nav_dict["name"] = child.text
            if id == "0x7C696E75":
                nav_dict["units"] = child.text
            if id == "0x7D6CD4DB":
                nav_array = np.fromstring(child.text.strip(), sep=" ")
                nav_size = nav_array.size
                if nav_size < 2:
                    has_nav = False
                else:
                    nav_dict["scale"] = nav_array[1] - nav_array[0]
                    nav_dict["offset"] = nav_array[0]
                    nav_dict["size"] = nav_size
                    nav_dict["navigate"] = True
        if has_nav:
            self.axes[tag] = nav_dict
        return has_nav, nav_size

    def _set_signal_axis(self, xml_element):
        """Helper method to extract signal axis information.

        Parameters
        ----------
        xml_element: xml.etree.ElementTree.Element
            Head level metadata element.
        tag: Str
            Axis name.
        """
        signal_dict = dict()
        signal_dict["navigate"] = False
        for child in xml_element:
            id = self._get_id(child)
            if id == "0x7D6CD4DB":
                signal_array = np.fromstring(child.text.strip(), sep=" ")
                if signal_array.size > 1:
                    if signal_array[0] > signal_array[1]:
                        signal_array = signal_array[::-1]
                        self._reverse_signal = True
                    else:
                        self._reverse_signal = False
                    if self._use_uniform_signal_axis:
                        offset, scale = np.polynomial.polynomial.polyfit(
                            np.arange(signal_array.size), signal_array, deg=1
                        )
                        signal_dict["offset"] = offset
                        signal_dict["scale"] = scale
                        signal_dict["size"] = signal_array.size
                        scale_compare = 100 * np.max(
                            np.abs(np.diff(signal_array) - scale) / scale
                        )
                        if scale_compare > 1:
                            _logger.warning(
                                f"The relative variation of the signal-axis-scale ({scale_compare:.2f}%) exceeds 1%.\n"
                                "                              "
                                "Using a non-uniform-axis is recommended."
                            )
                    else:
                        signal_dict["axis"] = signal_array
                else:  # pragma: no cover
                    if self._use_uniform_signal_axis:  # pragma: no cover
                        _logger.warning(  # pragma: no cover
                            "Signal only contains one entry.\n"  # pragma: no cover
                            "                              "  # pragma: no cover
                            "Using non-uniform-axis independent of use_uniform_signal_axis setting"  # pragma: no cover
                        )  # pragma: no cover
                    signal_dict["axis"] = signal_array  # pragma: no cover
            if id == "0x7C696E75":
                units = child.text
                if "/" in units and units[-3:] == "abs":
                    signal_dict["name"] = "Wavenumber"
                    signal_dict["units"] = units[:-4]
                elif "/" in units and units[-1] == "m":
                    signal_dict["name"] = "Raman Shift"
                    signal_dict["units"] = units
                elif units[-2:] == "eV":
                    signal_dict["name"] = "Energy"
                    signal_dict["units"] = units
                elif "/" not in units and units[-1] == "m":
                    signal_dict["name"] = "Wavelength"
                    signal_dict["units"] = units
                else:
                    _logger.warning(
                        "Cannot extract type of signal axis from units, using wavelength as name."
                    )  # pragma: no cover
                    signal_dict["name"] = "Wavelength"  # pragma: no cover
                    signal_dict["units"] = units  # pragma: no cover
        self.axes["signal_dict"] = signal_dict

    def _sort_nav_axes(self):
        """Sort the navigation/signal axes, such that (X, Y, Spectrum) = (1, 0, 2) (for map)
        or (X/Y, Spectrum) = (0, 1) or (Spectrum) = (0) (for linescan/spectrum).
        """
        self.axes["signal_dict"]["index_in_array"] = len(self.axes) - 1
        if self._has_nav2:
            self.axes["nav2_dict"]["index_in_array"] = 0
            if self._has_nav1:
                self.axes["nav1_dict"]["index_in_array"] = 1
        elif self._has_nav1 and not self._has_nav2:
            self.axes["nav1_dict"]["index_in_array"] = 0
        self.axes = sorted(self.axes.values(), key=lambda item: item["index_in_array"])

    def get_axes(self):
        """Extract navigation/signal axes data from file."""
        self.axes = dict()
        self._has_nav1 = False
        self._has_nav2 = False
        for child in self._axis_root:
            if self._get_id(child) == "0x0":
                self._set_signal_type(child)
            if self._get_id(child) == "0x1":
                self._set_signal_axis(child)
            if self._get_id(child) == "0x2":
                self._has_nav1, self._nav1_size = self._set_nav_axis(child, "nav1_dict")
            if self._get_id(child) == "0x3":
                self._has_nav2, self._nav2_size = self._set_nav_axis(child, "nav2_dict")

        self._sort_nav_axes()

    def get_data(self):
        """Extract data from file."""
        data_raw = self._lsx_matrix.findall("LSX_Row")
        ## lexicographical ordering -> 3x3 map -> 9 rows
        num_rows = len(data_raw)
        if num_rows == 0:
            _logger.critical("No data found.")  # pragma: no cover
        elif num_rows == 1:
            ## Spectrum
            self.data = np.fromstring(data_raw[0].text.strip(), sep=" ")
            if self._reverse_signal:
                self.data = self.data[::-1]
        else:
            ## linescan or map
            num_cols = self._get_size(data_raw[0])
            self.data = np.empty((num_rows, num_cols))
            for i, row in enumerate(data_raw):
                row_array = np.fromstring(row.text.strip(), sep=" ")
                if self._reverse_signal:
                    row_array = row_array[::-1]
                self.data[i, :] = row_array
            ## reshape the array (lexicographic -> cartesian)
            ## reshape depends on available axes
            if self._has_nav2:
                if self._has_nav1:
                    self.data = np.reshape(
                        self.data, (self._nav2_size, self._nav1_size, num_cols)
                    )
                else:
                    self.data = np.reshape(
                        self.data, (self._nav2_size, num_cols)
                    )  # pragma: no cover
            elif self._has_nav1 and not self._has_nav2:
                self.data = np.reshape(self.data, (self._nav1_size, num_cols))

    def map_metadata(self):
        """Maps original_metadata to metadata dictionary."""
        general = {}
        signal = {}
        laser = {"Filter": {}, "Polarizer": {}}
        spectrometer = {"Grating": {}, "Polarizer": {}}
        detector = {"processing": {}}
        spectral_image = {}
        sample = {}

        sample["description"] = self.original_metadata["file_information"].get("Sample")

        general["title"] = self._title
        general["original_filename"] = self._file_path.name
        general["notes"] = self.original_metadata["file_information"].get("Remark")
        try:
            date, time = self.original_metadata["date"]["Acquired"].split(" ")
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover
        else:
            general["date"] = date
            general["time"] = time

        signal["signal_type"] = self._signal_type
        signal["signal_dimension"] = 1
        try:
            intensity_axis = self.original_metadata["experimental_setup"]["signal type"]
            intensity_units = self.original_metadata["experimental_setup"][
                "signal units"
            ]
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover
        else:
            if intensity_axis == "Intens":
                intensity_axis = "Intensity"
            if intensity_units == "Cnt/sec":
                intensity_units = "Counts/s"
            if intensity_units == "Cnt":
                intensity_units = "Counts"
            signal["quantity"] = f"{intensity_axis} ({intensity_units})"

        laser["wavelength"] = self.original_metadata["experimental_setup"].get(
            "Laser (nm)"
        )
        laser["objective_magnification"] = self.original_metadata[
            "experimental_setup"
        ].get("Objective")
        laser["Filter"]["optical_density"] = self.original_metadata[
            "experimental_setup"
        ].get("ND Filter (%)")
        laser["Polarizer"]["polarizer_type"] = self.original_metadata[
            "experimental_setup"
        ].get("Laser. Pol.")
        laser["Polarizer"]["angle"] = self.original_metadata["experimental_setup"].get(
            "Laser Pol. (°)"
        )

        spectrometer["central_wavelength"] = self.original_metadata[
            "experimental_setup"
        ].get("Spectro (nm)")
        spectrometer["model"] = self.original_metadata["experimental_setup"].get(
            "Instrument"
        )
        spectrometer["Grating"]["groove_density"] = self.original_metadata[
            "experimental_setup"
        ].get("Grating (gr/mm)")
        spectrometer["entrance_slit_width"] = self.original_metadata[
            "experimental_setup"
        ].get("Hole")
        spectrometer["spectral_range"] = self.original_metadata[
            "experimental_setup"
        ].get("Range")
        spectrometer["Polarizer"]["polarizer_type"] = self.original_metadata[
            "experimental_setup"
        ].get("Raman. Pol.")
        spectrometer["Polarizer"]["angle"] = self.original_metadata[
            "experimental_setup"
        ].get("Raman Pol. (°)")

        detector["model"] = self.original_metadata["experimental_setup"].get("Detector")
        detector["delay_time"] = self.original_metadata["experimental_setup"].get(
            "Delay time (s)"
        )
        detector["binning"] = self.original_metadata["experimental_setup"].get(
            "Binning"
        )
        detector["temperature"] = self.original_metadata["experimental_setup"].get(
            "Detector temperature (°C)"
        )
        detector["exposure_per_frame"] = self.original_metadata[
            "experimental_setup"
        ].get("Acq. time (s)")
        detector["frames"] = self.original_metadata["experimental_setup"].get(
            "Accumulations"
        )
        detector["processing"]["autofocus"] = self.original_metadata[
            "experimental_setup"
        ].get("Autofocus")
        detector["processing"]["swift"] = self.original_metadata[
            "experimental_setup"
        ].get("SWIFT")
        detector["processing"]["auto_exposure"] = self.original_metadata[
            "experimental_setup"
        ].get("AutoExposure")
        detector["processing"]["spike_filter"] = self.original_metadata[
            "experimental_setup"
        ].get("Spike filter")
        detector["processing"]["de_noise"] = self.original_metadata[
            "experimental_setup"
        ].get("DeNoise")
        detector["processing"]["ics_correction"] = self.original_metadata[
            "experimental_setup"
        ].get("ICS correction")
        detector["processing"]["dark_correction"] = self.original_metadata[
            "experimental_setup"
        ].get("Dark correction")
        detector["processing"]["inst_process"] = self.original_metadata[
            "experimental_setup"
        ].get("Inst. Process")

        ## extra units here, because rad vs. deg
        if "rotation angle (rad)" in self.original_metadata["experimental_setup"]:
            spectral_image["rotation_angle"] = self.original_metadata[
                "experimental_setup"
            ]["rotation angle (rad)"] * (180 / np.pi)
            spectral_image["rotation_angle_units"] = "°"

        ## settings for glued spectra
        if "Windows" in self.original_metadata["experimental_setup"]:
            detector["glued_spectrum"] = True
            detector["glued_spectrum_windows"] = self.original_metadata[
                "experimental_setup"
            ]["Windows"]
        else:
            detector["glued_spectrum"] = False

        ## calculate and set integration time
        try:
            integration_time = (
                self.original_metadata["experimental_setup"]["Accumulations"]
                * self.original_metadata["experimental_setup"]["Acq. time (s)"]
            )
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover
        else:
            detector["integration_time"] = integration_time

        ## convert filter range from percentage (0-100) to (0-1)
        try:
            laser["Filter"]["optical_density"] /= 100
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover

        ## convert entrance_hole_width to mm
        try:
            spectrometer["entrance_slit_width"] /= 100
            spectrometer["pinhole"] = spectrometer["entrance_slit_width"]
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover

        self.metadata = {
            "General": general,
            "Signal": signal,
            "Sample": sample,
            "Acquisition_instrument": {
                "Laser": laser,
                "Spectrometer": spectrometer,
                "Detector": detector,
                "Spectral_image": spectral_image,
            },
        }

        _remove_none_from_dict(self.metadata)


def file_reader(filename, lazy=False, use_uniform_signal_axis=False):
    """
    Read data from .xml files saved using Horiba Jobin Yvon's LabSpec software.

    Parameters
    ----------
    %s
    %s
    use_uniform_signal_axis : bool, default=False
        Can be specified to choose between non-uniform or uniform signal axis.
        If ``True``, the ``scale`` attribute is calculated from the average delta
        along the signal axis and a warning is raised in case the delta varies
        by more than 1 percent.

    %s
    """
    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")
    if not isinstance(filename, Path):
        filename = Path(filename)
    jy = JobinYvonXMLReader(
        file_path=filename, use_uniform_signal_axis=use_uniform_signal_axis
    )
    jy.parse_file()
    jy.get_original_metadata()
    jy.get_axes()
    jy.get_data()
    jy.map_metadata()
    dictionary = {
        "data": jy.data,
        "axes": jy.axes,
        "metadata": deepcopy(jy.metadata),
        "original_metadata": deepcopy(jy.original_metadata),
    }
    return [
        dictionary,
    ]


file_reader.__doc__ %= (FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
import os
import struct
import warnings
import xml.etree.ElementTree as ET
from collections import OrderedDict
from glob import glob

import numpy as np
from dateutil import parser

from rsciio._docstrings import FILENAME_DOC, LAZY_DOC, RETURNS_DOC
from rsciio.utils._array import sarray2dict
from rsciio.utils._dictionary import DTBox

_logger = logging.getLogger(__name__)


data_types = {
    "1": "<u1",
    "2": "<u2",
    "3": "<u4",
    "4": "<i1",
    "5": "<i2",
    "6": "<i4",
    "7": "<f4",
    "8": "<f8",
    "9": "<c8",
    "10": "<c16",
}

XY_TAG_ID = 16706  # header contains XY calibration


def readLELongLong(file):
    """Read 8 bytes as *little endian* integer in file"""
    read_bytes = file.read(8)
    return struct.unpack("<Q", read_bytes)[0]


def readLELong(file):
    """Read 4 bytes as *little endian* integer in file"""
    read_bytes = file.read(4)
    return struct.unpack("<L", read_bytes)[0]


def readLEShort(file):
    """Read 2 bytes as *little endian* integer in file"""
    read_bytes = file.read(2)
    return struct.unpack("<H", read_bytes)[0]


def dimension_array_dtype(n, DescriptionLength, UnitsLength):
    dt_list = [
        ("Dim-%s_DimensionSize" % n, ("<u4")),
        ("Dim-%s_CalibrationOffset" % n, "<f8"),
        ("Dim-%s_CalibrationDelta" % n, "<f8"),
        ("Dim-%s_CalibrationElement" % n, "<u4"),
        ("Dim-%s_DescriptionLength" % n, "<u4"),
        ("Dim-%s_Description" % n, (bytes, DescriptionLength)),
        ("Dim-%s_UnitsLength" % n, "<u4"),
        ("Dim-%s_Units" % n, (bytes, UnitsLength)),
    ]
    return dt_list


def get_lengths(file):
    file.seek(24, 1)
    description_length = readLELong(file)
    file.seek(description_length, 1)
    unit_length = readLELong(file)
    file.seek(unit_length, 1)
    return description_length, unit_length


def get_header_dtype_list(file):
    # Read the first part of the header
    header_list1 = [
        ("ByteOrder", "<u2"),
        ("SeriesID", "<u2"),
        ("SeriesVersion", "<u2"),
        ("DataTypeID", "<u4"),
        ("TagTypeID", "<u4"),
        ("TotalNumberElements", "<u4"),
        ("ValidNumberElements", "<u4"),
    ]
    header1 = np.fromfile(file, dtype=np.dtype(header_list1), count=1)
    # Depending on the SeriesVersion, the OffsetArrayOffset is 4 or 8 bytes
    if header1["SeriesVersion"] <= 528:
        OffsetArrayOffset_dtype = "<u4"
        beginning_dimension_array_section = 30
    else:
        OffsetArrayOffset_dtype = "<u8"
        beginning_dimension_array_section = 34

    # Once we know the type of the OffsetArrayOffset, we can continue reading
    # the 2nd part of the header
    file.seek(22)
    header_list2 = [
        ("OffsetArrayOffset", OffsetArrayOffset_dtype),
        ("NumberDimensions", "<u4"),
    ]
    header2 = np.fromfile(file, dtype=np.dtype(header_list2), count=1)
    header_list = header_list1 + header_list2
    # Go to the beginning of the dimension array section
    file.seek(beginning_dimension_array_section)
    for n in range(1, header2["NumberDimensions"][0] + 1):
        description_length, unit_length = get_lengths(file)
        header_list += dimension_array_dtype(n, description_length, unit_length)
    file.seek(0)
    return header_list


def get_data_dtype_list(file, offset, record_by):
    if record_by == "spectrum":
        file.seek(offset + 20)
        data_type = readLEShort(file)
        array_size = readLELong(file)
        header = [
            ("CalibrationOffset", ("<f8")),
            ("CalibrationDelta", "<f8"),
            ("CalibrationElement", "<u4"),
            ("DataType", "<u2"),
            ("ArrayLength", "<u4"),
            ("Array", (data_types[str(data_type)], array_size)),
        ]
        shape = array_size
    elif record_by == "image":  # Untested
        file.seek(offset + 40)
        data_type = readLEShort(file)
        array_size_x = readLELong(file)
        array_size_y = readLELong(file)
        header = [
            ("CalibrationOffsetX", ("<f8")),
            ("CalibrationDeltaX", "<f8"),
            ("CalibrationElementX", "<u4"),
            ("CalibrationOffsetY", ("<f8")),
            ("CalibrationDeltaY", "<f8"),
            ("CalibrationElementY", "<u4"),
            ("DataType", "<u2"),
            ("ArraySizeX", "<u4"),
            ("ArraySizeY", "<u4"),
            ("Array", (data_types[str(data_type)], (array_size_x, array_size_y))),
        ]
        shape = (array_size_x, array_size_y)
    return header, shape


def get_data_tag_dtype_list(data_type_id):
    if data_type_id == XY_TAG_ID:
        header = [
            ("TagTypeID", ("<u2")),
            ("Unknown", ("<u2")),  # Not in Boothroyd description. = 0
            ("Time", "<u4"),  # The precision is one second...
            ("PositionX", "<f8"),
            ("PositionY", "<f8"),
        ]
    else:  # elif data_type_id == ?????, 16722?
        header = [
            ("TagTypeID", ("<u2")),
            # Not in Boothroyd description. = 0. Not tested.
            ("Unknown", ("<u2")),
            ("Time", "<u4"),  # The precision is one second...
        ]
    return header


def log_struct_array_values(struct_array):
    for key in struct_array.dtype.names:
        if (
            not isinstance(struct_array[key], np.ndarray)
            or np.array(struct_array[key].shape).sum() == 1
        ):
            _logger.info("%s : %s", key, struct_array[key])
        else:
            _logger.info("%s : Array", key)


def guess_record_by(record_by_id):
    if record_by_id == 16672:
        return "spectrum"
    else:
        return "image"


def parse_ExperimentalDescription(et, dictree):
    dictree.add_node(et.tag)
    dictree = dictree[et.tag]
    for data in et.find("Root").findall("Data"):
        label = data.find("Label").text
        value = data.find("Value").text
        units = data.find("Unit").text
        item = label if not units else label + "_%s" % units
        try:
            # try to coerce value to decimal representation
            value = float(value) if units else value
        except ValueError:
            _logger.warning(
                f"Expected decimal value for {label}, but received {value} instead"
            )
        dictree[item] = value


def parse_TrueImageHeaderInfo(et, dictree):
    dictree.add_node(et.tag)
    dictree = dictree[et.tag]
    et = ET.fromstring(et.text)
    for data in et.findall(b"Data"):
        dictree[data.find(b"Index").text] = float(data.find(b"Value").text)


def emixml2dtb(et, dictree):
    if et.tag == "ExperimentalDescription":
        parse_ExperimentalDescription(et, dictree)
        return
    elif et.tag == "TrueImageHeaderInfo":
        parse_TrueImageHeaderInfo(et, dictree)
        return
    if et.text:
        dictree[et.tag] = et.text
        return
    else:
        dictree.add_node(et.tag)
        for child in et:
            emixml2dtb(child, dictree[et.tag])


def emi_reader(filename, lazy=False, only_valid_data=True, dump_xml=False):
    # TODO: recover the tags from the emi file. It is easy: just look for
    # <ObjectInfo> and </ObjectInfo>. It is standard xml :)
    # xml chunks are identified using UUID, if we can find how these UUID are
    # generated then, it will possible to match to the corresponding ser file
    # and add the detector information in the metadata
    objects = get_xml_info_from_emi(filename)
    orig_fname = filename
    filename = os.path.splitext(filename)[0]
    if dump_xml:
        for i, obj in enumerate(objects):
            with open(filename + "-object-%s.xml" % i, "w") as f:
                f.write(obj)

    ser_files = sorted(glob(filename + "_[0-9].ser"))
    sers = []
    for f in ser_files:
        _logger.info("Opening %s", f)
        try:
            sers.extend(ser_reader(f, objects, lazy, only_valid_data))
        except IOError:  # Probably a single spectrum that we don't support
            continue

        index = int(os.path.splitext(f)[0].split("_")[-1]) - 1
        op = DTBox(sers[-1]["original_metadata"], box_dots=True)

        # defend against condition where more ser files are present than object
        # metadata defined in emi
        if index < len(objects):
            emixml2dtb(ET.fromstring(objects[index]), op)
        else:
            _logger.warning(
                f"{orig_fname} did not contain any metadata for "
                f"{f}, so only .ser header information was read"
            )
        sers[-1]["original_metadata"] = op.to_dict()
    return sers


def file_reader(filename, lazy=False, only_valid_data=True):
    """
    Read sets of ``.ser`` and ``.emi`` files from the FEI/ThermoFisher software TIA
    (TEM Imaging & Analysis).

    Parameters
    ----------
    %s
    %s
    only_valid_data : bool, Default=True
        For cases, where acquisition of series or linescan data stopped before
        the end. If `True`, load only the acquired data. If `False`, the empty
        data are filled with zeros.

    %s
    """
    ext = os.path.splitext(filename)[1][1:]
    if ext.lower() == "ser":
        to_return = ser_reader(
            filename, objects=None, lazy=lazy, only_valid_data=only_valid_data
        )
    elif ext.lower() == "emi":
        to_return = emi_reader(filename, lazy, only_valid_data)
    else:
        raise ValueError(f"'{ext}' is not a supported extension for the TIA reader.")

    return to_return


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def load_ser_file(filename):
    _logger.info("Opening the file: %s", filename)
    with open(filename, "rb") as f:
        header = np.fromfile(f, dtype=np.dtype(get_header_dtype_list(f)), count=1)
        _logger.info("Header info:")
        log_struct_array_values(header[0])

        if header["ValidNumberElements"] == 0:
            raise IOError(
                "The file does not contains valid data. "
                "If it is a single spectrum, the data is contained in the  "
                ".emi file but HyperSpy cannot currently extract this "
                "information."
            )

        # Read the first element of data offsets
        f.seek(header["OffsetArrayOffset"][0])
        # OffsetArrayOffset can contain 4 or 8 bytes integer depending if the
        # data have been acquired using a 32 or 64 bits platform.
        if header["SeriesVersion"] <= 528:
            data_offset = readLELong(f)
            data_offset_array = np.fromfile(
                f, dtype="<u4", count=header["ValidNumberElements"][0]
            )
        else:
            data_offset = readLELongLong(f)
            data_offset_array = np.fromfile(
                f, dtype="<u8", count=header["ValidNumberElements"][0]
            )
        data_dtype_list, shape = get_data_dtype_list(
            f, data_offset, guess_record_by(header["DataTypeID"])
        )
        tag_dtype_list = get_data_tag_dtype_list(header["TagTypeID"])
        f.seek(data_offset)
        data = np.empty(
            header["ValidNumberElements"][0],
            dtype=np.dtype(data_dtype_list + tag_dtype_list),
        )
        for i, offset in enumerate(data_offset_array):
            data[i] = np.fromfile(
                f, dtype=np.dtype(data_dtype_list + tag_dtype_list), count=1
            )
            f.seek(offset)
        _logger.info("Data info:")
        log_struct_array_values(data[0])
    return header, data


def get_xml_info_from_emi(emi_file):
    with open(emi_file, "rb") as f:
        tx = f.read()
    objects = []
    i_start = 0
    while i_start != -1:
        i_start += 1
        i_start = tx.find(b"<ObjectInfo>", i_start)
        i_end = tx.find(b"</ObjectInfo>", i_start)
        objects.append(tx[i_start : i_end + 13].decode("utf-8"))
    return objects[:-1]


def get_calibration_from_position(position):
    """Compute the size, scale and offset of a linear axis from coordinates.

    This function assumes rastering on a regular grid for the full size of
    each dimension before rastering over another one. Fox example: a11, a12,
    a13, a21, a22, a23 for a 2x3 grid.

    Parameters
    ----------
    position: numpy array.
        Position coordinates of the axis. Normally as in PositionX/Y of the
        ser file.

    Returns
    -------
    axis_attr: dictionary with `size`, `scale`, `offeset` keys.

    """
    offset = position[0]
    for i, x in enumerate(position):
        if x != position[0]:
            break
    if i == len(position) - 1:
        # No scanning over this axis
        scale = 0
        size = 0
    else:
        scale = x - position[0]
        if i == 1:  # Rastering over this dimension first
            for j, x in enumerate(position[1:]):
                if x == position[0]:
                    break
            size = j + 1
        else:  # Second rastering dimension
            size = len(position) / i
    axis_attr = {"size": size, "scale": scale, "offset": offset}
    return axis_attr


def get_axes_from_position(header, data):
    array_shape = []
    axes = []
    array_size = int(header["ValidNumberElements"])
    if data[b"TagTypeID"][0] == XY_TAG_ID:
        xcal = get_calibration_from_position(data[b"PositionX"])
        ycal = get_calibration_from_position(data[b"PositionY"])
        if xcal[b"size"] == 0 and ycal[b"size"] != 0:
            # Vertical line scan
            axes.append(
                {
                    "name": "x",
                    "units": "meters",
                    "index_in_array": 0,
                }
            )
            axes[-1].update(xcal)
            array_shape.append(axes[-1]["size"])

        elif xcal[b"size"] != 0 and ycal[b"size"] == 0:
            # Horizontal line scan
            axes.append(
                {
                    "name": "y",
                    "units": "meters",
                    "index_in_array": 0,
                }
            )
            axes[-1].update(ycal)
            array_shape.append(axes[-1]["size"])

        elif xcal[b"size"] * ycal[b"size"] == array_size:
            # Signal2D
            axes.append(
                {
                    "name": "y",
                    "units": "meters",
                    "index_in_array": 0,
                }
            )
            axes[-1].update(ycal)
            array_shape.append(axes[-1]["size"])
            axes.append(
                {
                    "name": "x",
                    "units": "meters",
                    "index_in_array": 1,
                }
            )
            axes[-1].update(xcal)
            array_shape.append(axes[-1]["size"])
        elif xcal[b"size"] == ycal[b"size"] == array_size:
            # Oblique line scan
            scale = np.sqrt(xcal["scale"] ** 2 + ycal["scale"] ** 2)
            axes.append(
                {
                    "name": "x",
                    "units": "meters",
                    "index_in_array": 0,
                    "offset": 0,
                    "scale": scale,
                    "size": xcal["size"],
                }
            )
            array_shape.append(axes[-1]["size"])
        else:
            raise IOError
    else:
        array_shape = [header["ValidNumberElements"]]
        axes.append(
            {
                "name": "Unknown dimension",
                "offset": 0,
                "scale": 1,
                "units": "",
                "size": header["ValidNumberElements"],
                "index_in_array": 0,
            }
        )
    return array_shape, axes


def convert_xml_to_dict(xml_object):
    op = DTBox(box_dots=True)
    emixml2dtb(ET.fromstring(xml_object), op)
    return op


def ser_reader(filename, objects=None, lazy=False, only_valid_data=True):
    """
    Reads the information from the file and returns it in the HyperSpy
    required format.
    """
    header, data = load_ser_file(filename)
    record_by = guess_record_by(header["DataTypeID"])
    ndim = int(header["NumberDimensions"][0])
    date, time = None, None
    if objects is not None:
        objects_dict = convert_xml_to_dict(objects[0])
        try:
            acq_date = objects_dict.ObjectInfo.AcquireDate
            date, time = _get_date_time(acq_date)
        except AttributeError:
            _logger.warning(
                f"AcquireDate not found in metadata of {filename};"
                " Not setting metadata date or time"
            )
    if (
        "PositionY" in data.dtype.names
        and len(data["PositionY"]) > 1
        and (data["PositionY"][0] == data["PositionY"][1])
    ):
        # The spatial dimensions are stored in F order i.e. X, Y, ...
        order = "F"
    else:
        # The spatial dimensions are stored in C order i.e. ..., Y, X
        order = "C"
    if ndim == 0 and header["ValidNumberElements"] != 0:
        # The calibration of the axes are not stored in the header.
        # We try to guess from the position coordinates.
        array_shape, axes = get_axes_from_position(header=header, data=data)
    else:
        axes = []
        array_shape = [
            None,
        ] * int(ndim)
        spatial_axes = ["x", "y"][:ndim]
        for i in range(ndim):
            idim = 1 + i if order == "C" else ndim - i
            if (
                record_by == "spectrum"
                or header["Dim-%i_DimensionSize" % (i + 1)][0] != 1
            ):
                units = (
                    header["Dim-%i_Units" % (idim)][0].decode("utf-8")
                    if header["Dim-%i_UnitsLength" % (idim)] > 0
                    else None
                )
                if units == "meters":
                    name = spatial_axes.pop() if order == "F" else spatial_axes.pop(-1)
                else:
                    name = None
                axes.append(
                    {
                        "offset": header["Dim-%i_CalibrationOffset" % idim][0],
                        "scale": header["Dim-%i_CalibrationDelta" % idim][0],
                        "units": units,
                        "size": header["Dim-%i_DimensionSize" % idim][0],
                        "name": name,
                        "navigate": True,
                    }
                )
                array_shape[i] = header["Dim-%i_DimensionSize" % idim][0]

        # Deal with issue when TotalNumberElements does not equal
        # ValidNumberElements for ndim==1.
        if (
            ndim == 1
            and (header["TotalNumberElements"] != header["ValidNumberElements"][0])
            and only_valid_data
        ):
            if header["ValidNumberElements"][0] == 1:
                # no need for navigation dimension
                array_shape = []
                axes = []
            else:
                array_shape[0] = header["ValidNumberElements"][0]
                axes[0]["size"] = header["ValidNumberElements"][0]

    # Spectral dimension
    if record_by == "spectrum":
        axes.append(
            {
                "offset": data["CalibrationOffset"][0],
                "scale": data["CalibrationDelta"][0],
                "size": data["ArrayLength"][0],
                "index_in_array": header["NumberDimensions"][0],
                "navigate": False,
            }
        )

        # FEI seems to use the international system of units (SI) for the
        # energy scale (eV).
        axes[-1]["units"] = "eV"
        axes[-1]["name"] = "Energy"

        array_shape.append(data["ArrayLength"][0])

    elif record_by == "image":
        if objects is not None:
            units = _guess_units_from_mode(objects_dict, header)
        else:
            units = "meters"
        # Y axis
        axes.append(
            {
                "name": "y",
                "offset": data["CalibrationOffsetY"][0]
                - data["CalibrationElementY"][0] * data["CalibrationDeltaY"][0],
                "scale": data["CalibrationDeltaY"][0],
                "units": units,
                "size": data["ArraySizeY"][0],
                "navigate": False,
            }
        )
        array_shape.append(data["ArraySizeY"][0])
        # X axis
        axes.append(
            {
                "name": "x",
                "offset": data["CalibrationOffsetX"][0]
                - data["CalibrationElementX"][0] * data["CalibrationDeltaX"][0],
                "scale": data["CalibrationDeltaX"][0],
                "size": data["ArraySizeX"][0],
                "units": units,
                "navigate": False,
            }
        )
        array_shape.append(data["ArraySizeX"][0])

    # FEI seems to use the international system of units (SI) for the
    # spatial scale. However, we prefer to work in nm
    for axis in axes:
        if axis["units"] == "meters":
            axis["units"] = "nm"
            axis["scale"] *= 10**9
        elif axis["units"] == "1/meters":
            axis["units"] = "1 / nm"
            axis["scale"] /= 10**9

    # Remove Nones from array_shape caused by squeezing size 1 dimensions
    array_shape = [dim for dim in array_shape if dim is not None]
    if lazy:
        from dask import delayed
        from dask.array import from_delayed

        val = delayed(load_only_data, pure=True)(
            filename, array_shape, record_by, len(axes), only_valid_data=only_valid_data
        )
        dc = from_delayed(val, shape=array_shape, dtype=data["Array"].dtype)
    else:
        dc = load_only_data(
            filename,
            array_shape,
            record_by,
            len(axes),
            data=data,
            header=header,
            only_valid_data=only_valid_data,
        )

    original_metadata = OrderedDict()
    header_parameters = sarray2dict(header)
    sarray2dict(data, header_parameters)
    # We remove the Array key to save memory avoiding duplication
    del header_parameters["Array"]
    original_metadata["ser_header_parameters"] = header_parameters
    metadata = {
        "General": {
            "original_filename": os.path.split(filename)[1],
        },
        "Signal": {
            "signal_type": "",
        },
    }
    if date is not None and time is not None:
        metadata["General"]["date"] = date
        metadata["General"]["time"] = time
    dictionary = {
        "data": dc,
        "metadata": metadata,
        "axes": axes,
        "original_metadata": original_metadata,
        "mapping": mapping,
    }
    return [
        dictionary,
    ]


def load_only_data(
    filename,
    array_shape,
    record_by,
    num_axes,
    data=None,
    header=None,
    only_valid_data=True,
):
    if data is None:
        header, data = load_ser_file(filename)
    # If the acquisition stops before finishing the job, the stored file will
    # report the requested size even though no values are recorded. Therefore
    # if the shapes of the retrieved array does not match that of the data
    # dimensions we must fill the rest with zeros or (better) nans if the
    # dtype is float
    if np.prod(array_shape) != np.prod(data["Array"].shape):
        if int(header["NumberDimensions"][0]) == 1 and only_valid_data:
            # No need to fill with zeros if `TotalNumberElements !=
            # ValidNumberElements` for series data.
            # The valid data is always `0:ValidNumberElements`
            dc = data["Array"][0 : header["ValidNumberElements"][0], ...]
            array_shape[0] = header["ValidNumberElements"][0]
        else:
            # Maps will need to be filled with zeros or nans
            dc = np.zeros(np.prod(array_shape), dtype=data["Array"].dtype)
            if dc.dtype is np.dtype("f") or dc.dtype is np.dtype("f8"):
                dc[:] = np.nan
            dc[: data["Array"].ravel().shape[0]] = data["Array"].ravel()
    else:
        dc = data["Array"]

    dc = dc.reshape(array_shape)
    if record_by == "image":
        dc = dc[..., ::-1, :]
    if num_axes != len(dc.shape):
        dc = dc.squeeze()
    if num_axes != len(dc.shape):
        raise IOError("Please report this issue to the HyperSpy developers.")
    return dc


def _guess_units_from_mode(objects_dict, header):
    # in case the xml file doesn't contain the "Mode" or the header doesn't
    # contain 'Dim-1_UnitsLength', return "meters" as default, which will be
    # OK most of the time
    warn_str = (
        "The navigation axes units could not be determined. "
        "Setting them to `nm`, but this may be wrong."
    )
    try:
        mode = objects_dict.ObjectInfo.ExperimentalDescription.Mode
        isCamera = "CameraNamePath" in objects_dict.ObjectInfo.AcquireInfo.keys()
    except AttributeError:  # in case the xml chunk doesn't contain the Mode
        warnings.warn(warn_str)
        return "meters"  # Most of the time, the unit will be meters!
    if "Dim-1_UnitsLength" in header.dtype.fields:
        # assuming that for an image stack, the UnitsLength of the "3rd"
        # dimension is 0
        isImageStack = header["Dim-1_UnitsLength"][0] == 0
        # Workaround: if this is not an image stack and not a STEM image, then
        # we assume that it should be a diffraction
        isDiffractionScan = header["Dim-1_DimensionSize"][0] > 1 and not isImageStack
    else:
        warnings.warn(warn_str)
        return "meters"  # Most of the time, the unit will be meters!

    _logger.info(objects_dict.ObjectInfo.AcquireInfo)
    _logger.info("mode: %s", mode)
    _logger.info("isCamera: %s", isCamera)
    _logger.info("isImageStack: %s", isImageStack)
    _logger.info("isImageStack: %s", isDiffractionScan)
    if "STEM" in mode:
        # data recorded in STEM with a camera, so we assume, it's a diffraction
        # in case we can't make use the detector is a camera, use a workaround
        if isCamera or isDiffractionScan:
            return "1/meters"
        else:
            return "meters"
    elif "Diffraction" in mode:
        return "1/meters"
    else:
        return "meters"


def _get_simplified_mode(mode):
    if "STEM" in mode:
        return "STEM"
    else:
        return "TEM"


def _get_date_time(value):
    dt = parser.parse(value)
    return dt.date().isoformat(), dt.time().isoformat()


def _get_microscope_name(value):
    return value.replace("Microscope ", "")


mapping = {
    "ObjectInfo.ExperimentalDescription.High_tension_kV": (
        "Acquisition_instrument.TEM.beam_energy",
        None,
    ),
    "ObjectInfo.ExperimentalDescription.Microscope": (
        "Acquisition_instrument.TEM.microscope",
        _get_microscope_name,
    ),
    "ObjectInfo.ExperimentalDescription.Mode": (
        "Acquisition_instrument.TEM.acquisition_mode",
        _get_simplified_mode,
    ),
    "ObjectInfo.ExperimentalDescription.Camera length_m": (
        "Acquisition_instrument.TEM.camera_length",
        lambda x: x * 1e3,
    ),
    "ObjectInfo.ExperimentalDescription.Magnification_x": (
        "Acquisition_instrument.TEM.magnification",
        None,
    ),
    "ObjectInfo.AcquireInfo.CameraNamePath": (
        "Acquisition_instrument.TEM.Detector.Camera.Name",
        None,
    ),
    "ObjectInfo.AcquireInfo.DwellTimePath": (
        "Acquisition_instrument.TEM.Detector.Camera.exposure",
        None,
    ),
    "ObjectInfo.ExperimentalDescription.Stage_A_deg": (
        "Acquisition_instrument.TEM.Stage.tilt_alpha",
        None,
    ),
    "ObjectInfo.ExperimentalDescription.Stage_B_deg": (
        "Acquisition_instrument.TEM.Stage.tilt_beta",
        None,
    ),
    "ObjectInfo.ExperimentalDescription.Stage_X_um": (
        "Acquisition_instrument.TEM.Stage.x",
        lambda x: x * 1e-3,
    ),
    "ObjectInfo.ExperimentalDescription.Stage_Y_um": (
        "Acquisition_instrument.TEM.Stage.y",
        lambda x: x * 1e-3,
    ),
    "ObjectInfo.ExperimentalDescription.Stage_Z_um": (
        "Acquisition_instrument.TEM.Stage.z",
        lambda x: x * 1e-3,
    ),
    "ObjectInfo.ExperimentalDescription.User": ("General.authors", None),
}



# -*- coding: utf-8 -*-
# Copyright 2024-2026 Delmic
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

from pathlib import Path

import numpy as np
import pytest

try:
    import hyperspy.api as hs
except ImportError:
    pytest.skip("hyperspy not installed", allow_module_level=True)

try:
    import lumispy
except ImportError:
    lumispy = None

pytest.importorskip("h5py", reason="h5py not installed")

testfile_dir = (Path(__file__).parent / "data" / "delmic").resolve()
testfile_intensity_path = (testfile_dir / "sparc-intensity.h5").resolve()
testfile_intensity_data_survey_path = (
    testfile_dir / "sparc-intensity-data-survey.npy"
).resolve()
testfile_intensity_drift_path = (testfile_dir / "sparc-intensity-drift.h5").resolve()
testfile_hyperspectral_path = (testfile_dir / "sparc-hyperspectral.h5").resolve()
testfile_hyperspectral_data_path = (
    testfile_dir / "sparc-hyperspectral-data.npy"
).resolve()
testfile_hyperspectral_data_survey_path = (
    testfile_dir / "sparc-hyperspectral-data-survey.npy"
).resolve()
testfile_hyperspectral_wavelengths_path = (
    testfile_dir / "sparc-hyperspectral-wavelengths.npy"
).resolve()
testfile_hyperspectral_spot_path = (
    testfile_dir / "sparc-hyperspectral-spot.h5"
).resolve()
testfile_temporaltrace_path = (testfile_dir / "sparc-time-correlator.h5").resolve()
testfile_temporaltrace_data_path = (
    testfile_dir / "sparc-time-correlator-data.npy"
).resolve()
testfile_temporaltrace_data_survey_path = (
    testfile_dir / "sparc-time-correlator-data-survey.npy"
).resolve()
testfile_temporaltrace_timelist_path = (
    testfile_dir / "sparc-time-correlator-timelist.npy"
).resolve()
testfile_temporaltrace_spot_path = (
    testfile_dir / "sparc-time-correlator-spot.h5"
).resolve()
testfile_streakcamera_path = (testfile_dir / "sparc-streak-camera.h5").resolve()
testfile_streakcamera_data_path = (
    testfile_dir / "sparc-streak-camera-data.npy"
).resolve()
testfile_streakcamera_data_survey_path = (
    testfile_dir / "sparc-streak-camera-data-survey.npy"
).resolve()
testfile_streakcamera_timelist_path = (
    testfile_dir / "sparc-streak-camera-timelist.npy"
).resolve()
testfile_streakcamera_wavelengths_path = (
    testfile_dir / "sparc-streak-camera-wavelengths.npy"
).resolve()
testfile_streakcamera_spot_path = (
    testfile_dir / "sparc-streak-camera-spot.h5"
).resolve()
testfile_ek_path = (testfile_dir / "sparc-e-k.h5").resolve()
testfile_ek_data_path = (testfile_dir / "sparc-e-k-data.npy").resolve()
testfile_ek_data_survey_path = (testfile_dir / "sparc-e-k-data-survey.npy").resolve()
testfile_ek_channels_path = (testfile_dir / "sparc-e-k-channels.npy").resolve()
testfile_ek_wavelengths_path = (testfile_dir / "sparc-e-k-wavelengths.npy").resolve()
testfile_ek_spot_path = (testfile_dir / "sparc-e-k-spot.h5").resolve()
testfile_AR_path = (testfile_dir / "sparc-angle-resolved.h5").resolve()
testfile_AR_data_path = (testfile_dir / "sparc-angle-resolved-data.npy").resolve()
testfile_AR_data_survey_path = (
    testfile_dir / "sparc-angle-resolved-data-survey.npy"
).resolve()
testfile_AR_angles_path = (testfile_dir / "sparc-angle-resolved-angles.npy").resolve()
testfile_AR_channels_path = (
    testfile_dir / "sparc-angle-resolved-channels.npy"
).resolve()
testfile_ar_pol_spot_path = (
    testfile_dir / "sparc-angle-resolved-pol-spot.h5"
).resolve()

testfile_hspy_path = (
    Path(__file__).parent / "data" / "hspy" / "example1_v1.2.hdf5"
).resolve()
testfile_arina_path = (
    Path(__file__).parent / "data" / "arina" / "test_00.h5"
).resolve()


# Intensity dataset
def test_read_data_intensity():
    """Test reading data for a CL intensity dataset."""
    s = hs.load(testfile_intensity_path, reader="Delmic", signal="all")

    assert len(s) == 3
    assert s[0].metadata.General.title == "CL intensity"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[-1].metadata.General.title == "Secondary electrons survey"


def test_read_data_intensity_CL():
    """Test reading data for a CL intensity dataset."""
    s = hs.load(testfile_intensity_path, reader="Delmic")  # Default signal is "cl"
    assert isinstance(s, hs.signals.BaseSignal)

    # Expect 2x2 pixels
    x = np.array([27147, 28907])
    y = np.array([26964, 27695])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)

    # test X axis
    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate

    np.testing.assert_allclose(s.axes_manager[0].scale, 200)
    np.testing.assert_allclose(s.axes_manager[0].offset, -100)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    # test Y axis
    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate

    np.testing.assert_allclose(s.axes_manager[1].scale, -200)
    np.testing.assert_allclose(s.axes_manager[1].offset, 100)
    np.testing.assert_allclose(s.axes_manager[1].size, 2)

    # Metadata
    assert s.metadata["Signal"]["quantity"] == "Counts"

    # Original metadata
    np.testing.assert_almost_equal(s.original_metadata.Magnification, 555555.55555555)
    assert s.original_metadata.SVIData.Company == "Delmic"


def test_read_data_intensity_SE():
    """Test reading data for a CL intensity dataset."""
    s = hs.load(testfile_intensity_path, reader="Delmic", signal="se")
    assert isinstance(s, hs.signals.BaseSignal)

    x = np.array([27308, 28592])
    y = np.array([27851, 27958])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)

    # test X axis
    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate

    np.testing.assert_allclose(s.axes_manager[0].scale, 200)
    np.testing.assert_allclose(s.axes_manager[0].offset, -100)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    # test Y axis
    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate

    np.testing.assert_allclose(s.axes_manager[1].scale, -200)
    np.testing.assert_allclose(s.axes_manager[1].offset, 100)
    np.testing.assert_allclose(s.axes_manager[1].size, 2)

    # Metadata
    assert s.metadata["Signal"]["quantity"] == "Counts"

    # Original metadata
    np.testing.assert_almost_equal(s.original_metadata.Magnification, 555555.55555555)


def test_read_data_intensity_survey():
    """Test reading data for a CL intensity dataset."""
    s = hs.load(testfile_intensity_path, reader="Delmic", signal="survey")
    assert isinstance(s, hs.signals.BaseSignal)

    # 256x256 px
    data = np.load(testfile_intensity_data_survey_path)
    np.testing.assert_allclose(s.data, data)

    # test X axis
    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate

    np.testing.assert_allclose(s.axes_manager[0].scale, 1.7578125)
    np.testing.assert_allclose(s.axes_manager[0].offset, -224.121094)
    np.testing.assert_allclose(s.axes_manager[0].size, 256)

    # test Y axis
    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate

    np.testing.assert_allclose(s.axes_manager[1].scale, -1.7578125)
    np.testing.assert_allclose(s.axes_manager[1].offset, 224.121094)
    np.testing.assert_allclose(s.axes_manager[1].size, 256)

    assert s.metadata["Signal"]["quantity"] == "Counts"

    np.testing.assert_almost_equal(s.original_metadata.Magnification, 555555.55555555)


def test_read_data_intensity_drift():
    """Test reading data for a CL intensity dataset with drift correction."""
    s = hs.load(testfile_intensity_drift_path, reader="Delmic", signal="all")

    assert len(s) == 4
    assert s[0].metadata.General.title == "CL intensity"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Anchor region"
    assert s[3].metadata.General.title == "Secondary electrons survey"

    anchor_data = s[2]
    assert anchor_data.data.shape == (134, 122, 4)


# Hyperspectral dataset
def test_read_data_hyperspectral():
    """Test reading data for a CL hyperspectral dataset."""
    s = hs.load(testfile_hyperspectral_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title.startswith("Spectrum")
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_hyperspectral_spot():
    """Test reading data for a CL hyperspectral dataset."""
    s = hs.load(testfile_hyperspectral_spot_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title.startswith("Spectrum")
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_hyperspectral_CL():
    """Test reading data for a CL hyperspectral dataset."""
    s = hs.load(testfile_hyperspectral_path, reader="Delmic", signal="cl")
    if lumispy:
        assert isinstance(s, lumispy.signals.CLSEMSpectrum)
    else:
        assert isinstance(s, hs.signals.Signal1D)

    data = np.load(testfile_hyperspectral_data_path)
    np.testing.assert_allclose(s.data, data)

    # Axes
    ref = np.load(testfile_hyperspectral_wavelengths_path)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate

    np.testing.assert_allclose(s.axes_manager[0].scale, 963.862901465797)
    np.testing.assert_allclose(s.axes_manager[0].offset, -893.135842)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -963.862901465797)
    np.testing.assert_allclose(s.axes_manager[1].offset, 1326.3523814)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)

    assert s.axes_manager[2].name == "Wavelength"
    assert s.axes_manager[2].units == "nm"
    assert not s.axes_manager[2].navigate
    np.testing.assert_allclose(s.axes_manager[2].axis, ref)

    # Metadata
    assert s.metadata["Signal"]["quantity"] == "Counts"

    # Original metadata
    assert isinstance(s.original_metadata.AcquisitionDate, float)


def test_read_data_hyperspectral_spot_CL():
    """Test reading data for a CL hyperspectral dataset."""
    s = hs.load(testfile_hyperspectral_spot_path, reader="Delmic", signal="cl")

    np.testing.assert_allclose(s.data.shape, 335)

    assert s.axes_manager[0].name == "Wavelength"
    assert s.axes_manager[0].units == "nm"
    assert not s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].size, 335)


def test_read_data_hyperspectral_SE():
    """Test reading data for a CL hyperspectral dataset."""
    s = hs.load(testfile_hyperspectral_path, reader="Delmic", signal="se")
    assert isinstance(s, hs.signals.BaseSignal)

    x = np.array([27265, 27598])
    y = np.array([27892, 28124])
    z = np.array([28299, 28468])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)
    np.testing.assert_allclose(s.data[2], z)

    # test X axis
    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 963.862901465797)
    np.testing.assert_allclose(s.axes_manager[0].offset, -893.135842)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -963.862901465797)
    np.testing.assert_allclose(s.axes_manager[1].offset, 1326.3523814)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)

    # Metadata
    assert s.metadata["Signal"]["quantity"] == "Counts"

    # Original metadata
    assert s.original_metadata.Magnification == 10000.0


def test_read_data_hyperspectral_survey():
    """Test reading data for a CL hyperspectral dataset."""
    s = hs.load(testfile_hyperspectral_path, reader="Delmic", signal="survey")
    assert isinstance(s, hs.signals.BaseSignal)
    data = np.load(testfile_hyperspectral_data_survey_path)

    np.testing.assert_allclose(s.data, data)

    # test X axis
    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 27.343750000000004)
    np.testing.assert_allclose(s.axes_manager[0].offset, -6986.328125)
    np.testing.assert_allclose(s.axes_manager[0].size, 512)

    # test Y axis
    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -27.343750000000004)
    np.testing.assert_allclose(s.axes_manager[1].offset, 6986.328125)
    np.testing.assert_allclose(s.axes_manager[1].size, 512)

    # Metadata
    assert s.metadata["Signal"]["quantity"] == "Counts"

    # Original metadata
    assert s.original_metadata.Magnification == 10000.0


# Time-resolved dataset
def test_read_data_temporaltrace():
    """Test reading data for a CL decay trace or g(2) datasets."""
    s = hs.load(testfile_temporaltrace_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "Time Correlator"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_temporaltrace_spot():
    """Test reading data for a CL decay trace or g(2) datasets."""
    s = hs.load(testfile_temporaltrace_spot_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "Time Correlator"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_temporaltrace_CL():
    """Test reading data for a CL decay trace or g(2) datasets."""
    s = hs.load(testfile_temporaltrace_path, reader="Delmic", signal="cl")
    if lumispy:
        assert isinstance(s, lumispy.signals.LumiTransient)
    else:
        assert isinstance(s, hs.signals.Signal1D)

    data = np.load(testfile_temporaltrace_data_path)
    np.testing.assert_allclose(s.data, data)

    ref = np.load(testfile_temporaltrace_timelist_path)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 917.1372563963987)
    np.testing.assert_allclose(s.axes_manager[0].offset, -773.568628)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -917.1372563963987)
    np.testing.assert_allclose(s.axes_manager[1].offset, 1267.1372563964)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)

    assert s.axes_manager[2].name == "Time"
    assert s.axes_manager[2].units == "ns"
    assert not s.axes_manager[2].navigate
    np.testing.assert_allclose(s.axes_manager[2].axis, ref)

    assert s.metadata.General.title == "Time Correlator"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert isinstance(s.original_metadata.AcquisitionDate, float)


def test_read_data_temporaltrace_CL_spot():
    """Test reading data for a CL decay trace or g(2) datasets."""
    s = hs.load(testfile_temporaltrace_spot_path, reader="Delmic", signal="cl")
    if lumispy:
        assert isinstance(s, lumispy.signals.LumiTransient)
    else:
        assert isinstance(s, hs.signals.Signal1D)

    np.testing.assert_allclose(s.data.shape, 65536)

    np.testing.assert_allclose(s.axes_manager[0].size, 65536)
    assert s.axes_manager[0].name == "Time"
    assert s.axes_manager[0].units == "ns"
    assert not s.axes_manager[0].navigate


def test_read_data_temporaltrace_SE():
    """Test reading data for a CL decay trace or g(2) datasets."""
    s = hs.load(testfile_temporaltrace_path, reader="Delmic", signal="se")
    assert isinstance(s, hs.signals.BaseSignal)

    x = np.array([32766, 32766])
    y = np.array([32766, 32766])
    z = np.array([32766, 32766])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)
    np.testing.assert_allclose(s.data[2], z)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 917.1372563963987)
    np.testing.assert_allclose(s.axes_manager[0].offset, -773.568628)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -917.1372563963987)
    np.testing.assert_allclose(s.axes_manager[1].offset, 1267.1372563964)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)

    assert s.metadata.General.title == "Secondary electrons concurrent"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


def test_read_data_temporaltrace_survey():
    """Test reading data for a CL decay trace or g(2) datasets."""
    s = hs.load(testfile_temporaltrace_path, reader="Delmic", signal="survey")
    assert isinstance(s, hs.signals.BaseSignal)

    data = np.load(testfile_temporaltrace_data_survey_path)
    np.testing.assert_allclose(s.data, data)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 27.343750000000004)
    np.testing.assert_allclose(s.axes_manager[0].offset, -6986.328125)
    np.testing.assert_allclose(s.axes_manager[0].size, 512)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -27.343750000000004)
    np.testing.assert_allclose(s.axes_manager[1].offset, 6986.328125)
    np.testing.assert_allclose(s.axes_manager[1].size, 512)

    assert s.metadata.General.title == "Secondary electrons survey"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


# Streak camera dataset
def test_read_data_streakcamera():
    """Test reading data for a CL streak camera dataset."""
    s = hs.load(testfile_streakcamera_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "Temporal Spectrum"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_streakcamera_spot():
    """Test reading data for a CL streak camera dataset."""
    s = hs.load(testfile_streakcamera_spot_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "Temporal Spectrum"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_streakcamera_CL():
    """Test reading data for a CL streak camera dataset."""
    s = hs.load(testfile_streakcamera_path, reader="Delmic", signal="cl")
    if lumispy:
        assert isinstance(s, lumispy.signals.LumiTransientSpectrum)
    else:
        assert isinstance(s, hs.signals.Signal2D)

    data = np.load(testfile_streakcamera_data_path)
    np.testing.assert_allclose(s.data, data)

    # Axes
    ref_t = np.load(testfile_streakcamera_timelist_path)
    ref_w = np.load(testfile_streakcamera_wavelengths_path)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "µm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 3.10737984559)
    np.testing.assert_allclose(s.axes_manager[0].offset, -5.518681)
    np.testing.assert_allclose(s.axes_manager[0].size, 3)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "µm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -3.10737984559)
    np.testing.assert_allclose(s.axes_manager[1].offset, 2.12726981)
    np.testing.assert_allclose(s.axes_manager[1].size, 2)

    assert s.axes_manager[2].name == "Wavelength"
    assert s.axes_manager[2].units == "nm"
    assert not s.axes_manager[2].navigate
    np.testing.assert_allclose(s.axes_manager[2].axis, ref_w)

    assert s.axes_manager[3].name == "Time"
    assert s.axes_manager[3].units == "ns"
    assert not s.axes_manager[3].navigate
    np.testing.assert_allclose(s.axes_manager[3].axis, ref_t)

    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert isinstance(s.original_metadata.AcquisitionDate, float)


def test_read_data_streakcamera_CL_spot():
    """Test reading data for a CL streak camera dataset."""
    s = hs.load(testfile_streakcamera_spot_path, reader="Delmic", signal="cl")

    np.testing.assert_allclose(s.data.shape, 256)

    # Axes
    np.testing.assert_allclose(s.axes_manager[0].size, 256)
    assert s.axes_manager[0].name == "Wavelength"
    assert s.axes_manager[0].units == "nm"
    assert not s.axes_manager[0].navigate

    np.testing.assert_allclose(s.axes_manager[1].size, 256)
    assert s.axes_manager[1].name == "Time"
    assert s.axes_manager[1].units == "ns"
    assert not s.axes_manager[1].navigate


def test_read_data_streakcamera_SE():
    """Test reading data for a CL streak camera dataset."""
    s = hs.load(testfile_streakcamera_path, reader="Delmic", signal="se")
    assert isinstance(s, hs.signals.BaseSignal)

    x = np.array([5443, 90, 5318])
    y = np.array([241, 5256, 174])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)

    # Axes
    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "µm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 3.10737984559)
    np.testing.assert_allclose(s.axes_manager[0].offset, -5.518681)
    np.testing.assert_allclose(s.axes_manager[0].size, 3)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "µm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -3.10737984559)
    np.testing.assert_allclose(s.axes_manager[1].offset, 2.12726981)
    np.testing.assert_allclose(s.axes_manager[1].size, 2)

    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


def test_read_data_streakcamera_survey():
    """Test reading data for a CL streak camera dataset."""
    s = hs.load(testfile_streakcamera_path, reader="Delmic", signal="survey")

    data = np.load(testfile_streakcamera_data_survey_path)
    np.testing.assert_allclose(s.data, data)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 48.828125)
    np.testing.assert_allclose(s.axes_manager[0].offset, -12475.585938)
    np.testing.assert_allclose(s.axes_manager[0].size, 512)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -48.828125)
    np.testing.assert_allclose(s.axes_manager[1].offset, 12475.585938)
    np.testing.assert_allclose(s.axes_manager[1].size, 512)

    assert s.metadata["General"]["title"] == "Secondary electrons survey"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


# E-k dataset
def test_read_data_ek():
    """Test reading data for a CL AR Spectrum (E-k) dataset."""
    s = hs.load(testfile_ek_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "AR Spectrum"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_ek_spot():
    """Test reading data for a CL AR Spectrum (E-k) dataset."""
    s = hs.load(testfile_ek_spot_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "AR Spectrum"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_ek_CL():
    """Test reading data for a CL AR Spectrum (E-k) dataset."""
    s = hs.load(testfile_ek_path, reader="Delmic", signal="cl")
    assert isinstance(s, hs.signals.Signal2D)

    data = np.load(testfile_ek_data_path)
    np.testing.assert_allclose(s.data, data)

    ref_a = np.load(testfile_ek_channels_path)
    ref_w = np.load(testfile_ek_wavelengths_path)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 299.0099783670874)
    np.testing.assert_allclose(s.axes_manager[0].offset, -483.594683)
    np.testing.assert_allclose(s.axes_manager[0].size, 3)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -299.0099783670874)
    np.testing.assert_allclose(s.axes_manager[1].offset, 232.865286374114)
    np.testing.assert_allclose(s.axes_manager[1].size, 2)

    assert s.axes_manager[2].name == "Wavelength"
    assert s.axes_manager[2].units == "nm"
    assert not s.axes_manager[2].navigate
    np.testing.assert_allclose(s.axes_manager[2].axis, ref_w)

    assert s.axes_manager[3].name == "Angle"
    assert s.axes_manager[3].units == ""
    assert not s.axes_manager[3].navigate
    np.testing.assert_allclose(s.axes_manager[3].axis, ref_a)

    assert s.metadata.General.title == "AR Spectrum"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.PolePosition == [15.25, 18.125]


def test_read_data_ek_CL_spot():
    """Test reading data for a CL AR Spectrum (E-k) dataset."""
    s = hs.load(testfile_ek_spot_path, reader="Delmic", signal="cl")

    assert s.data.shape == (270, 320)

    np.testing.assert_allclose(s.axes_manager[0].size, 320)
    assert s.axes_manager[0].name == "Wavelength"
    assert s.axes_manager[0].units == "nm"
    assert not s.axes_manager[0].navigate

    np.testing.assert_allclose(s.axes_manager[1].size, 270)
    assert s.axes_manager[1].name == "Angle"
    assert s.axes_manager[1].units == ""
    assert not s.axes_manager[1].navigate


def test_read_data_ek_SE():
    """Test reading data for a CL AR Spectrum (E-k) dataset."""
    s = hs.load(testfile_ek_path, reader="Delmic", signal="se")
    assert isinstance(s, hs.signals.BaseSignal)

    x = np.array([32766, 32766, 32766])
    y = np.array([32766, 32766, 32766])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 299.0099783670874)
    np.testing.assert_allclose(s.axes_manager[0].offset, -483.594683)
    np.testing.assert_allclose(s.axes_manager[0].size, 3)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -299.0099783670874)
    np.testing.assert_allclose(s.axes_manager[1].offset, 232.865286374114)
    np.testing.assert_allclose(s.axes_manager[1].size, 2)

    assert s.metadata.General.title == "Secondary electrons concurrent"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


def test_read_data_ek_survey():
    """Test reading data for a CL AR Spectrum (E-k) dataset."""
    s = hs.load(testfile_ek_path, reader="Delmic", signal="survey")

    data = np.load(testfile_ek_data_survey_path)
    np.testing.assert_allclose(s.data, data)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 48.828125)
    np.testing.assert_allclose(s.axes_manager[0].offset, -12475.585938)
    np.testing.assert_allclose(s.axes_manager[0].size, 512)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -48.828125)
    np.testing.assert_allclose(s.axes_manager[1].offset, 12475.585938)
    np.testing.assert_allclose(s.axes_manager[1].size, 512)

    assert s.metadata.General.title == "Secondary electrons survey"
    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


# Angle-resolved dataset
def test_read_data_AR():
    """Test reading data for a CL AR dataset."""
    s = hs.load(testfile_AR_path, reader="Delmic", signal="all")
    assert len(s) == 3
    assert s[0].metadata.General.title == "Angle-resolved"
    assert s[1].metadata.General.title == "Secondary electrons concurrent"
    assert s[2].metadata.General.title == "Secondary electrons survey"


def test_read_data_AR_CL():
    """Test reading data for a CL AR dataset."""
    s = hs.load(testfile_AR_path, reader="Delmic", signal="cl")
    assert isinstance(s, hs.signals.Signal2D)

    data = np.load(testfile_AR_data_path)
    np.testing.assert_allclose(s.data, data)

    ref_a = np.load(testfile_AR_angles_path)
    ref_b = np.load(testfile_AR_channels_path)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 963.8629014657938)
    np.testing.assert_allclose(s.axes_manager[0].offset, -885.49392)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -963.8629014657942)
    np.testing.assert_allclose(s.axes_manager[1].offset, 1356.92001)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)

    assert s.axes_manager[2].name == "Angle A"
    assert s.axes_manager[2].units == ""
    assert not s.axes_manager[2].navigate
    np.testing.assert_allclose(s.axes_manager[2].axis, ref_a)

    assert s.axes_manager[3].name == "Angle B"
    assert s.axes_manager[3].units == ""
    assert not s.axes_manager[3].navigate
    np.testing.assert_allclose(s.axes_manager[3].axis, ref_b)

    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.PolePosition == [114.5, 129.75]


def test_read_data_AR_SE():
    """Test reading data for a CL AR dataset."""
    s = hs.load(testfile_AR_path, reader="Delmic", signal="se")
    assert isinstance(s, hs.signals.BaseSignal)

    x = np.array([32932, 33065])
    y = np.array([33203, 32495])
    z = np.array([32431, 32565])
    np.testing.assert_allclose(s.data[0], x)
    np.testing.assert_allclose(s.data[1], y)
    np.testing.assert_allclose(s.data[2], z)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 963.8629014657938)
    np.testing.assert_allclose(s.axes_manager[0].offset, -885.49392)
    np.testing.assert_allclose(s.axes_manager[0].size, 2)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -963.8629014657942)
    np.testing.assert_allclose(s.axes_manager[1].offset, 1356.92001)
    np.testing.assert_allclose(s.axes_manager[1].size, 3)

    assert s.metadata["Signal"]["quantity"] == "Counts"

    assert s.original_metadata.Magnification == 10000.0


def test_read_data_AR_survey():
    """Test reading data for a CL AR dataset."""
    s = hs.load(testfile_AR_path, reader="Delmic", signal="survey")
    assert isinstance(s, hs.signals.BaseSignal)

    data = np.load(testfile_AR_data_survey_path)
    np.testing.assert_allclose(s.data, data)

    assert s.axes_manager[0].name == "X"
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[0].navigate
    np.testing.assert_allclose(s.axes_manager[0].scale, 27.343750000000004)
    np.testing.assert_allclose(s.axes_manager[0].offset, -6986.328125)
    np.testing.assert_allclose(s.axes_manager[0].size, 512)

    assert s.axes_manager[1].name == "Y"
    assert s.axes_manager[1].units == "nm"
    assert s.axes_manager[1].navigate
    np.testing.assert_allclose(s.axes_manager[1].scale, -27.343750000000004)
    np.testing.assert_allclose(s.axes_manager[1].offset, 6986.328125)
    np.testing.assert_allclose(s.axes_manager[1].size, 512)

    assert s.metadata["Signal"]["quantity"] == "Counts"
    assert s.metadata["General"]["title"] == "Secondary electrons survey"

    assert s.original_metadata.Magnification == 10000.0


def test_read_data_ar_pol():
    """Test reading data for a CL AR polarized dataset."""
    s = hs.load(testfile_ar_pol_spot_path, reader="Delmic", signal="all")
    assert len(s) == 8

    # We expect 6 AR images, each with a different polarizations
    polarizations = set()
    for d in s[:6]:
        assert d.metadata.General.title == "Angle-resolved"
        assert d.data.shape == (128, 128)
        pol = d.metadata.Acquisition_instrument.Spectrometer.Filter.position
        polarizations.add(pol)

    # If set is not of length 6, then some polarizations are identical
    assert len(polarizations) == 6

    assert s[-2].metadata.General.title == "Secondary electrons concurrent"
    assert s[-1].metadata.General.title == "Secondary electrons survey"


def test_wrong_arguments():
    """
    Attempt to load an HDF5 file with wrong arguments should raise a ValueError
    """
    with pytest.raises(ValueError):
        hs.load(testfile_intensity_path, reader="Delmic", signal="wrong_signal")

    with pytest.raises(NotImplementedError):
        hs.load(testfile_hyperspectral_path, reader="Delmic", signal="cl", lazy=True)


def test_wrong_format():
    """
    Attempt to load an HDF5 file not of the correct format should raise an IOError
    """
    with pytest.raises(IOError):
        hs.load(testfile_arina_path, reader="Delmic", signal="all")

    with pytest.raises(IOError):
        hs.load(testfile_hspy_path, reader="Delmic", signal="all")



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

from box import Box


class DTBox(Box):
    """
    Subclass of Box to help migration from hyperspy `DictionaryTreeBrowser`
    to `Box` when splitting IO code from hyperspy to rosettasciio.

    When using `box_dots=True`, by default, period will be removed from keys.
    To support period containing keys, use `box_dots=False, default_box=True`.
    https://github.com/cdgriffith/Box/wiki/Types-of-Boxes#default-box
    """

    def add_node(self, path):
        keys = path.split(".")
        for key in keys:
            if self.get(key) is None:
                self[key] = {}
            self = self[key]

    def set_item(self, path, value):
        if self.get(path) is None:
            self.add_node(path)
        self[path] = value

    def has_item(self, path):
        return self.get(path) is not None


def dump_dictionary(
    file, dic, string="root", node_separator=".", value_separator=" = "
):
    for key in list(dic.keys()):
        if isinstance(dic[key], dict):
            dump_dictionary(file, dic[key], string + node_separator + key)
        else:
            file.write(
                string + node_separator + key + value_separator + str(dic[key]) + "\n"
            )
